import re
import zipfile
from pathlib import Path
from typing import Literal, get_args, TypedDict
from urllib.request import urlopen

import numpy as np
import requests
import tensorflow as tf
import tensorflow.keras
from ruamel.yaml import YAML
from scipy.ndimage import binary_dilation, gaussian_filter

from neuralib.io import CASCADE_MODEL_CACHE_DIRECTORY
from neuralib.typing import PathLike
from neuralib.util.utils import ensure_dir
from neuralib.util.verbose import fprint

__all__ = [
    'CASCADE_MODEL_TYPE',
    'CascadeModelConfig',
    'cascade_predict',
    'CascadeSpikePrediction'
]

CASCADE_MODEL_TYPE = Literal[
    'Global_EXC_1Hz_smoothing500ms',
    'Global_EXC_1Hz_smoothing1000ms',
    'Zebrafish_1Hz_smoothing1000ms',
    'Global_EXC_2Hz_smoothing300ms',
    'Global_EXC_2Hz_smoothing500ms',
    'Global_EXC_2Hz_smoothing1000ms',
    'Global_EXC_2.5Hz_smoothing400ms_high_noise',
    'Global_EXC_3Hz_smoothing400ms',
    'Global_EXC_3Hz_smoothing400ms_high_noise',
    'Global_EXC_3Hz_smoothing400ms_causalkernel',
    'Global_EXC_4.25Hz_smoothing300ms',
    'Global_EXC_4.25Hz_smoothing300ms_high_noise',
    'Global_EXC_4.25Hz_smoothing300ms_causalkernel',
    'Global_EXC_5Hz_smoothing200ms',
    'Global_EXC_5Hz_smoothing200ms_causalkernel',
    'Global_EXC_6Hz_smoothing200ms',
    'Global_EXC_6Hz_smoothing200ms_causalkernel',
    'Global_EXC_7Hz_smoothing200ms',
    'Global_EXC_7Hz_smoothing200ms_causalkernel',
    'Global_EXC_7.5Hz_smoothing200ms_high_noise',
    'Global_EXC_7.5Hz_smoothing200ms',
    'Global_EXC_7.5Hz_smoothing200ms_causalkernel',
    'OGB_zf_pDp_7.5Hz_smoothing200ms',
    'OGB_zf_pDp_7.5Hz_smoothing200ms_causalkernel',
    'Global_EXC_10Hz_smoothing50ms',
    'Global_EXC_10Hz_smoothing50ms_causalkernel',
    'Global_EXC_10Hz_smoothing100ms',
    'Global_EXC_10Hz_smoothing100ms_causalkernel',
    'Global_EXC_10Hz_smoothing200ms',
    'Global_EXC_10Hz_smoothing200ms_causalkernel',
    'Global_EXC_12.5Hz_smoothing100ms',
    'Global_EXC_12.5Hz_smoothing100ms_causalkernel',
    'Global_EXC_12.5Hz_smoothing200ms',
    'Global_EXC_12.5Hz_smoothing200ms_causalkernel',
    'Global_EXC_15Hz_smoothing50ms',
    'Global_EXC_15Hz_smoothing50ms_causalkernel',
    'Global_EXC_15Hz_smoothing100ms_high_noise',
    'Global_EXC_15Hz_smoothing100ms',
    'Global_EXC_15Hz_smoothing100ms_causalkernel',
    'Global_EXC_15Hz_smoothing200ms',
    'Global_EXC_15Hz_smoothing200ms_causalkernel',
    'Global_INH_15Hz_smoothing100ms',
    'Global_EXC_17.5Hz_smoothing100ms',
    'Global_EXC_17.5Hz_smoothing100ms_causalkernel',
    'Global_EXC_17.5Hz_smoothing200ms',
    'Global_EXC_17.5Hz_smoothing200ms_causalkernel',
    'Global_EXC_20Hz_smoothing100ms',
    'Global_EXC_20Hz_smoothing100ms_causalkernel',
    'Global_EXC_20Hz_smoothing200ms',
    'Global_EXC_20Hz_smoothing200ms_causalkernel',
    'Global_EXC_25Hz_smoothing100ms',
    'Global_EXC_25Hz_smoothing100ms_causalkernel',
    'Global_EXC_25Hz_smoothing50ms',
    'Global_EXC_25Hz_smoothing50ms_causalkernel',
    'Global_EXC_30Hz_smoothing25ms',
    'Global_EXC_30Hz_smoothing25ms_causalkernel',
    'Global_EXC_30Hz_smoothing50ms',
    'Global_EXC_30Hz_smoothing50ms_high_noise',
    'Global_EXC_30Hz_smoothing50ms_causalkernel',
    'Global_EXC_30Hz_smoothing100ms',
    'Global_EXC_30Hz_smoothing100ms_causalkernel',
    'Global_EXC_30Hz_smoothing200ms',
    'Global_EXC_30Hz_smoothing100ms_causalkernel_high_noise',
    'Global_EXC_30Hz_smoothing100ms_high_noise',
    'Global_EXC_30Hz_smoothing200ms_causalkernel_high_noise',
    'Global_EXC_40Hz_smoothing25ms_causalkernel',
    'Global_EXC_40Hz_smoothing25ms',
    'Global_EXC_40Hz_smoothing25ms_high_noise',
    'Global_EXC_40Hz_smoothing50ms',
    'Global_EXC_40Hz_smoothing50ms_high_noise',
    'Global_EXC_40Hz_smoothing50ms_causalkernel',
    'Global_INH_30Hz_smoothing50ms',
    'Global_INH_30Hz_smoothing100ms',
    'Global_EXC_30Hz_smoothing50ms_asymmetric_window_1_frame',
    'Global_EXC_30Hz_smoothing50ms_asymmetric_window_2_frames',
    'Global_EXC_30Hz_smoothing50ms_asymmetric_window_4_frames',
    'Global_EXC_30Hz_smoothing50ms_asymmetric_window_6_frames',
    'Global_EXC_30Hz_smoothing50ms_asymmetric_window_8_frames',
    'GCaMP6f_mouse_30Hz_smoothing200ms',
    'Spinal_cord_excitatory_30Hz_smoothing50ms',
    'Spinal_cord_inhibitory_30Hz_smoothing50ms',
    'Spinal_cord_excitatory_3Hz_smoothing400ms_high_noise',
    'Spinal_cord_inhibitory_3Hz_smoothing400ms_high_noise',
    'Spinal_cord_excitatory_2.5Hz_smoothing400ms',
    'Spinal_cord_inhibitory_2.5Hz_smoothing400ms',
    'GC8_EXC_5Hz_smoothing400ms_high_noise',
    'GC8_EXC_5Hz_smoothing800ms_high_noise',
    'GC8_EXC_7.5Hz_smoothing100ms_high_noise',
    'GC8_EXC_7.5Hz_smoothing200ms_high_noise',
    'GC8_EXC_10Hz_smoothing150ms_high_noise',
    'GC8_EXC_10Hz_smoothing75ms_high_noise',
    'GC8_EXC_15Hz_smoothing100ms_high_noise',
    'GC8_EXC_15Hz_smoothing50ms_high_noise',
    'GC8_EXC_30Hz_smoothing25ms_high_noise',
    'GC8_EXC_30Hz_smoothing50ms_high_noise',
    'GC8_EXC_40Hz_smoothing15ms_high_noise',
    'GC8_EXC_40Hz_smoothing30ms_high_noise'
]


def cascade_predict(dff: np.ndarray,
                    model_type: CASCADE_MODEL_TYPE,
                    *,
                    threshold: int | bool = 0,
                    padding: float = 0,
                    verbose: bool = True,
                    chunks_mode_limit: float = 10,
                    cache_dir: PathLike | None = None) -> np.ndarray:
    """
    Spike prediction using Cascade pretrained model

    :param dff: dF/F activity to be predicted. `Array[float, [N, F]|F]`
    :param model_type: ``MODEL_TYPE``
    :param threshold: Allowed values: 0, 1 or False.
        0: All negative values are set to 0.
        1 or True: Threshold signal to set every signal which is smaller than the expected signal size of an action potential to zero (with dilated mask)
        False: No thresholding. The result can contain negative values as well
    :param padding: Value which is inserted for datapoints, where no prediction can be made
        (because of window around timepoint of prediction).
        Default value: np.nan, another recommended value would be 0 which circumvents some problems with following analysis.
    :param verbose: Verbose of model information
    :param chunks_mode_limit: Decrease the number if memory issue dealing with large input arrays.
    :param cache_dir: Cache directory for saving the model. If None, then used default under `~/.cache/neuralib`
    :return: Spiking probability as predicted by the model. `Array[float, [N, F]|F]`
    """
    cascade = CascadeSpikePrediction(
        dff,
        model_type,
        threshold=threshold,
        padding=padding,
        verbose=verbose,
        chunks_mode_limit=chunks_mode_limit,
        cache_dir=cache_dir
    )

    spike = cascade.run_spike_prediction()
    return spike


class CascadeModelConfig(TypedDict, total=False):
    model_name: str
    """Name of the model"""

    sampling_rate: int
    """Sampling rate in Hz"""

    training_datasets: list[str]
    """Dataset of ground truth data (in folder 'Ground_truth')"""

    placeholder_1: int
    """protect formatting"""

    noise_levels: list[int]
    """Noise levels for training (integers, normally 1-9)"""

    placeholder_2: int
    """protect formatting"""

    smoothing: float
    """Standard deviation of Gaussian smoothing in time (sec)"""

    causal_kernel: int
    """Smoothing kernel is symmetric in time (0) or is causal (1)"""

    windowsize: int
    """Windowsize in timepoints"""

    before_frac: float
    """Fraction of timepoints before prediction point (0-1)"""

    filter_sizes: list[int]
    """Filter sizes for each convolutional layer"""

    filter_numbers: list[int]
    """Filter numbers for each convolutional layer"""

    dense_expansion: int
    """For dense layer"""

    loss_function: str
    """gradient-descent loss function"""

    optimizer: str
    """Adagrad"""

    nr_of_epochs: int
    """Number of training epochs per model"""

    ensemble_size: int
    """Number of models trained for one noise level"""

    batch_size: int
    """Batch size"""

    training_finished: Literal['Yes', 'No', 'Running']
    """Yes / No / Running"""

    verbose: int
    """level of status messages (0: minimal, 1: standard, 2: most, 3: all)"""


class CascadeSpikePrediction:

    def __init__(
            self,
            dff: np.ndarray,
            model_type: CASCADE_MODEL_TYPE,
            *,
            threshold: int | bool = 0,
            padding: float = 0,
            verbose: bool = True,
            chunks_mode_limit: float = 10,
            cache_dir: PathLike | None = None
    ):
        """

        :param dff: dF/F activity to be predicted. `Array[float, [N, F]|F]`
        :param model_type: ``MODEL_TYPE``
        :param threshold: Allowed values: 0, 1 or False.
                0: All negative values are set to 0.
                1 or True: Threshold signal to set every signal which is smaller than the expected signal size of an action potential to zero (with dilated mask)
                False: No thresholding. The result can contain negative values as well
        :param padding: Value which is inserted for datapoints, where no prediction can be made (because of window around timepoint of prediction).
            Default value: np.nan, another recommended value would be 0 which circumvents some problems with following analysis.
        :param verbose: Verbose of model information
        :param chunks_mode_limit: Decrease the number if memory issue dealing with large input arrays.
        :param cache_dir: Cache directory for saving the model. If None, then used default under `~/.cache/neuralib`
        """
        self.dff = dff
        self.model_type = model_type

        if cache_dir is not None:
            cache_dir = Path(cache_dir)
        self.cache_dir = cache_dir or CASCADE_MODEL_CACHE_DIRECTORY

        # model io
        if not self.available_model_yaml.exists():
            self._download_model_yaml()
        self._check_instance()
        if not self.model_dir.exists():
            self._download_model()

        # predict instance
        self.threshold = threshold
        self.padding = padding
        self.verbose = verbose
        self.chunks_mode_limit = chunks_mode_limit

    def _download_model_yaml(self):
        url = 'https://raw.githubusercontent.com/HelmchenLabSoftware/Cascade/master/Pretrained_models/available_models.yaml'
        response = requests.get(url)

        if response.status_code == 200:
            ensure_dir(self.cache_dir)
            out = self.available_model_yaml
            with open(out, 'wb') as file:
                file.write(response.content)
                fprint(f'YAML file saved successfully as {out}', vtype='io')
        else:
            raise RuntimeError(f'Failed to download the file. Status code: {response.status_code}')

    def _check_instance(self):
        available_models = self.get_available_models()
        if self.model_type not in available_models:
            raise ValueError(f'{self.model_type} should be one of the {available_models}')

        # check update
        if set(available_models) != set(get_args(CASCADE_MODEL_TYPE)):
            fprint('update MODEL_TYPE, new model released', vtype='warning')

    def _download_model(self):
        link = self.model_link
        with urlopen(link) as response:
            data = response.read()

        tmp_file = self.cache_dir / 'tmp_zipped_model.zip'
        with open(tmp_file, 'wb') as f:
            f.write(data)

        with zipfile.ZipFile(tmp_file, "r") as zip_ref:
            zip_ref.extractall(path=self.model_dir)

        tmp_file.unlink()

        fprint(f'Pretrained model was saved in folder {self.model_dir}', vtype='io')

    # ============== #
    # All Model Info #
    # ============== #

    @property
    def available_model_yaml(self) -> Path:
        """All models link/info yaml"""
        return self.cache_dir / 'available_models.yaml'

    def get_available_models(self) -> list[CASCADE_MODEL_TYPE]:
        """Get all the available in :attr:`available_model_yaml`"""
        content = YAML().load(self.available_model_yaml)
        models = list(content.keys())
        return models

    # ============== #
    # A Model Config #
    # ============== #

    @property
    def model_link(self) -> str:
        """Link of specified model"""
        with open(self.available_model_yaml, 'r') as file:
            config = YAML().load(file)
        return config[self.model_type]['Link']

    @property
    def model_dir(self) -> Path:
        """Directory of specified model"""
        return self.cache_dir / self.model_type

    @property
    def config_file(self):
        """Config filepath of specified model"""
        return self.model_dir / 'config.yaml'

    def get_config(self) -> CascadeModelConfig:
        """``ModelConfig`` of specified model"""
        with open(self.config_file, 'r') as file:
            return YAML().load(file)

    # ================ #
    # Spike Prediction #
    # ================ #

    def run_spike_prediction(self) -> np.ndarray:
        """
        Spike prediction

        :return: Spiking probability as predicted by the model. `Array[float, [N, F]|F]`
        """
        total_array_size = self.dff.itemsize * self.dff.size * 64 / 1e9

        if total_array_size < self.chunks_mode_limit:
            spike = self._predict(self.dff, self.threshold, self.padding, self.verbose)
        else:
            n_neurons = self.dff.shape[0]
            n_frames = self.dff.shape[1]
            spike = np.zeros((n_neurons, n_frames))
            nb_chunks = int(np.ceil(total_array_size / 10))
            chunks = np.array_split(range(n_neurons), nb_chunks)
            for part in range(nb_chunks):
                part_dff = self.dff[chunks[part], :]
                spike[chunks[part], :] = self._predict(part_dff, self.threshold, self.padding, self.verbose)

        return spike

    def _predict(self, dff, threshold=0, padding: float = np.nan, verbose: bool = True):
        # reshape if only a single neuron's activity is provided
        if len(dff.shape) == 1:
            dff = np.expand_dims(dff, 0)

        cfg = self.get_config()
        verbose = cfg["verbose"]
        training_data = cfg["training_datasets"]
        ensemble_size = cfg["ensemble_size"]
        batch_size = cfg["batch_size"]
        sampling_rate = cfg["sampling_rate"]
        before_frac = cfg["before_frac"]
        window_size = cfg["windowsize"]
        noise_levels_model = cfg["noise_levels"]
        smoothing = cfg["smoothing"]
        causal_kernel = cfg["causal_kernel"]

        # calculate noise levels for each trace
        trace_noise_levels = calculate_noise_levels(dff, sampling_rate)

        # Get model paths as dictionary (key: noise_level) with lists of model path
        model_dict = get_model_paths(self.model_dir)

        if verbose:
            msg = (f'The selected model was trained on {len(training_data)} datasets, '
                   f'with {ensemble_size} ensembles for each noise level, at a sampling rate of {sampling_rate} Hz')

            if causal_kernel:
                msg += ", with a resampled ground truth that was smoothed with a causal kernel"
            else:
                msg += ", with a resampled ground truth that was smoothed with a Gaussian kernel"

            msg += f'of a standard deviation of {str(int(1000 * smoothing))} ms.'
            fprint(msg)

            fprint(f'Loaded model was trained at frame rate {sampling_rate} Hz')
            fprint(f'Given argument traces contains {dff.shape[0]} neurons and {dff.shape[1]} frames.')

            noise_mean = str(int(np.nanmean(trace_noise_levels * 100)) / 100)
            noise_std = str(int(np.nanstd(trace_noise_levels * 100)) / 100)
            fprint(f'Noise levels (mean, std; in standard units): {noise_mean}, {noise_std}')

        # XX has shape: (neurons, timepoints, windows)
        XX = preprocess_traces(dff, before_frac=before_frac, window_size=window_size)
        Y_predict = np.zeros((XX.shape[0], XX.shape[1]))

        # Compute difference of noise levels between each neuron and each model; find the best fit
        differences = np.array(trace_noise_levels)[:, None] - np.array(noise_levels_model)[None, :]
        relative_differences = np.min(differences, axis=1)
        if np.mean(relative_differences) > 2:
            fprint(
                f"WARNING: The available models cannot match the experimentally obtained noise levels (difference: {str(np.mean(relative_differences))}),"
                f"Please check that the computation of dF/F is performed correctly. Otherwise, please reach out and ask for pretrained models with higher noise level models "
                f"(see: https://github.com/HelmchenLabSoftware/Cascade/issues/61).", vtype='warning'
            )
        best_model_for_each_neuron = np.argmin(np.abs(differences), axis=1)

        # Use for each noise level the matching model
        for i, noise_level in enumerate(noise_levels_model):

            # select neurons which have this noise level:
            neuron_idx = np.where(best_model_for_each_neuron == i)[0]
            if verbose:
                print(f'\nPredictions for noise level {noise_level}')
            if len(neuron_idx) == 0:  # no neurons were selected
                if verbose:
                    print(f"\tNo neurons for this noise level: {noise_level}")
                continue  # jump to next noise level

            # load keras models for the given noise level
            models = []
            for model_path in model_dict[noise_level]:
                models.append(tf.keras.models.load_model(model_path))

            # select neurons and merge neurons and timepoints into one dimension
            XX_sel = XX[neuron_idx, :, :]

            XX_sel = np.reshape(XX_sel, (XX_sel.shape[0] * XX_sel.shape[1], XX_sel.shape[2]))
            XX_sel = np.expand_dims(XX_sel, axis=2)  # add empty third dimension to match training shape

            for j, model in enumerate(models):
                if verbose:
                    print("\t... ensemble", j)

                prediction_flat = model.predict(XX_sel, batch_size, verbose=verbose)
                prediction = np.reshape(prediction_flat, (len(neuron_idx), XX.shape[1]))
                Y_predict[neuron_idx, :] += prediction / len(models)  # average predictions

            # remove models from memory
            tensorflow.keras.backend.clear_session()

        # handle threshold
        if threshold is False:
            fprint('Skipping the thresholding. There can be negative values in the result.', vtype='warning')

        elif threshold == 1:  # or True
            # Cut off noise floor (lower than 1/e of a single action potential)
            # find out empirically  how large a single AP is (depends on frame rate and smoothing)
            single_spike = np.zeros(1001)
            single_spike[501] = 1
            single_spike_smoothed = gaussian_filter(single_spike.astype(float), sigma=smoothing * sampling_rate)
            threshold_value = np.max(single_spike_smoothed) / np.exp(1)

            # Set everything below threshold to zero.
            # Use binary dilation to avoid clipping of true events.
            for neuron in range(Y_predict.shape[0]):
                # ignore warning because of nan's in Y_predict in comparison with value
                with np.errstate(invalid="ignore"):
                    activity_mask = Y_predict[neuron, :] > threshold_value

                activity_mask = binary_dilation(activity_mask, iterations=int(smoothing * sampling_rate))
                Y_predict[neuron, ~activity_mask] = 0
                Y_predict[Y_predict < 0] = 0  # set possible negative values in dilated mask to 0

        elif threshold == 0:
            # ignore warning because of nan's in Y_predict in comparison with value
            with np.errstate(invalid="ignore"):
                Y_predict[Y_predict < 0] = 0

        else:
            raise ValueError(f'Invalid value of threshold "{threshold}". Only 0, 1 (or True) or False allowed')

        # NaN or 0 for first and last datapoints, for which no predictions can be made
        Y_predict[:, 0: int(before_frac * window_size)] = padding
        Y_predict[:, -int((1 - before_frac) * window_size):] = padding

        return Y_predict


def preprocess_traces(dff: np.ndarray,
                      before_frac: float,
                      window_size: int) -> np.ndarray:
    """
    Transform dF/F data into a format that can be used by the deep network.
    For each time point, a window of the size 'window_size' of the dF/F is extracted.

    :param dff: `Array[float, [N, F]]`
    :param before_frac: positioning of the window around the current time point; 0.5 means center position
    :param window_size: size of the receptive window of the deep network
    :return: a matrix with `Array[float, [N, F, W]]`

    """
    start = int(before_frac * window_size - 1)
    end = dff.shape[1] - window_size + start + 1

    # extract a moving window from the calcium trace
    window_indexes = (
            np.expand_dims(np.arange(window_size), 0) +
            np.expand_dims(np.arange(dff.shape[1] - window_size + 1), 0).T
    )

    X = np.full((*dff.shape, window_size), np.nan)
    X[:, start:end, :] = dff[:, window_indexes]

    return X


def get_model_paths(model_path: Path) -> dict[int, list[Path]]:
    """Find all models in the model folder and return as dictionary"""
    all_models = sorted(list(model_path.glob('*.h5')))
    if len(all_models) == 0:
        raise FileNotFoundError(f'No models (*.h5 files) were found in the specified folder "{model_path}".')

    # dictionary with key for noise level, entries are lists of models
    model_dict = {}
    for model_path in all_models:
        noise_level = int(re.findall("_NoiseLevel_(\d+)", str(model_path))[0])
        if noise_level not in model_dict:
            model_dict[noise_level] = list()

        model_dict[noise_level].append(model_path)

    return model_dict


def calculate_noise_levels(dff: np.ndarray, frame_rate: float) -> np.ndarray:
    """
    Computes the noise levels for each neuron of the input matrix 'dF_traces'.

    The noise level is computed as the median absolute dF/F difference
    between two subsequent time points. This is a outlier-robust measurement
    that converges to the simple standard deviation of the dF/F trace for
    uncorrelated and outlier-free dF/F traces.

    Afterwards, the value is divided by the square root of the frame rate
    in order to make it comparable across recordings with different frame rates.

    :param dff: Fluorescence changed dF/F. `Array[float, [N, F]]`
    :param frame_rate: frame rate
    :return: vector of noise levels for all neurons
    """
    noise_levels = np.nanmedian(np.abs(np.diff(dff, axis=-1)), axis=-1) / np.sqrt(frame_rate)
    return noise_levels * 100  # scale noise levels to percent



from typing import Literal, ClassVar

import matplotlib.colorbar
import matplotlib.colors as mcolors
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from neuralib.plot import plot_figure, ax_merge
from neuralib.typing import ArrayLike
from neuralib.typing import PathLike

__all__ = ['DotPlot']


class DotPlot:
    DEFAULT_COLOR_TITLE: ClassVar[str] = 'value'
    DEFAULT_SIZE_TITLE: ClassVar[str] = 'value'
    DEFAULT_MAX_MARKER_SIZE: ClassVar[float] = 500.0
    DEFAULT_MARKER_LEGEND_NUM: ClassVar[int] = 5

    def __init__(
            self,
            xlabel: ArrayLike,
            ylabel: ArrayLike,
            values: np.ndarray,
            *,
            scale: Literal['area', 'radius'] = 'radius',
            max_marker_size: float | None = None,
            size_title: str | None = None,
            size_legend_num: int | None = None,
            size_legend_as_int: bool = True,
            with_color: bool = False,
            cmap: mcolors.Colormap = 'Reds',
            colorbar_title: str | None = None,
            norm: mcolors.Normalize | None = None,
            cbar_vmin: float | None = None,
            cbar_vmax: float | None = None,
            figure_title: str | None = None,
            figure_output: PathLike | None = None,
            ax: Axes | None = None
    ):
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.values = values

        self._check_instance()

        # size
        self.scale = scale
        self.max_marker_size = max_marker_size or DotPlot.DEFAULT_MAX_MARKER_SIZE
        self.size_title = size_title or DotPlot.DEFAULT_SIZE_TITLE
        self.size_legend_num = size_legend_num or DotPlot.DEFAULT_MARKER_LEGEND_NUM
        self.size_as_int = size_legend_as_int

        # color
        self.with_color = with_color
        self.colorbar_title = colorbar_title or DotPlot.DEFAULT_COLOR_TITLE
        self.cmap = cmap
        self.cbar_vmin = cbar_vmin or np.min(values)
        self.cbar_vmax = cbar_vmax or np.max(values)
        self.norm = norm or self._default_norm()

        # figure
        self.figure_title = figure_title
        self.figure_output = figure_output
        self.ax = ax

    def _check_instance(self):
        x = np.array(self.xlabel)
        y = np.array(self.ylabel)

        if x.ndim != 1 or y.ndim != 1:
            raise ValueError('xlabel and ylabel must to be 1d')

        if not np.issubdtype(x.dtype, np.str_) or not np.issubdtype(y.dtype, np.str_):
            raise TypeError('xlabel and ylabel must be string array')

        nx = len(self.xlabel)
        ny = len(self.ylabel)

        if self.values.shape != (nx, ny):
            raise ValueError('shape inconsistent')

    def _default_norm(self) -> mcolors.Normalize:
        """default cbar norm"""
        return mcolors.Normalize(vmin=self.cbar_vmin, vmax=self.cbar_vmax)

    @property
    def scaling_factor(self) -> float:
        max_value = np.max(self.values)
        if self.scale == 'radius':
            max_value **= 2
        return self.max_marker_size / max_value

    @property
    def size(self) -> np.ndarray:
        """map value to size, which maximal ``MAX_SCATTER_SIZE``"""
        values = np.array([self.values]).flatten()
        if self.scale == 'radius':
            return (values ** 2) * self.scaling_factor
        elif self.scale == 'area':
            return values * self.scaling_factor
        else:
            raise ValueError(f'{self.scale}')

    def plot(self, **kwargs):
        values = np.array([self.values]).flatten()
        x, y = np.meshgrid(self.xlabel, self.ylabel, indexing='ij')

        if self.ax is None:
            self._plot_figure(x, y, values, **kwargs)
        else:
            self._plot_ax(x, y, values, **kwargs)

    def _plot_figure(self, x, y, values, **kwargs):
        """Plot in figure level"""
        with plot_figure(self.figure_output, 7, 4, figsize=(8, 5)) as _ax:
            ax = ax_merge(_ax)[:, :3]
            if self.with_color:
                ax.scatter(x.ravel(), y.ravel(),
                           s=self.size,
                           c=values,
                           cmap=self.cmap,
                           clip_on=False,
                           norm=self.norm,
                           edgecolor='gray',
                           **kwargs)

                ax_cbar = _ax[6, 3]
                self._plot_colorbar(ax_cbar, self.norm)
            else:
                ax.scatter(x.ravel(), y.ravel(),
                           s=self.size,
                           c='k',
                           clip_on=False,
                           **kwargs)
                ax_cbar = _ax[6, 3]
                ax_cbar.axis('off')

            ax_size = ax_merge(_ax)[:6, 3]
            ax_size.axis('off')
            self._plot_size_legend(ax_size)

            if self.figure_title is not None:
                ax.set_title(self.figure_title)

    def _plot_ax(self, x, y, values, **kwargs):
        """Plot wit existing Axes"""
        if self.with_color:
            im = self.ax.scatter(x.ravel(), y.ravel(),
                                 s=self.size,
                                 c=values,
                                 cmap=self.cmap,
                                 clip_on=False,
                                 norm=self.norm,
                                 edgecolor='gray',
                                 **kwargs)

            cbar = self.ax.figure.colorbar(im)
            cbar.ax.set_ylabel(self.colorbar_title)

        else:
            self.ax.scatter(x.ravel(), y.ravel(),
                            s=self.size,
                            c='k',
                            clip_on=False,
                            **kwargs)

        if self.figure_title is not None:
            self.ax.set_title(self.figure_title)

        self._plot_size_legend(self.ax)

    def _plot_size_legend(self, size_ax: Axes):
        values = np.round(
            np.linspace(self.norm.vmin, self.norm.vmax, num=self.size_legend_num), 2
        )

        if self.size_as_int:
            values = np.round(values).astype(np.int_)

        for v in values:
            if self.scale == 'radius':
                size = (v ** 2) * self.scaling_factor
            elif self.scale == 'area':
                size = v * self.scaling_factor
            else:
                raise ValueError('')

            size_ax.scatter([], [], s=size, c='k', label=str(v))

        handle, label = size_ax.get_legend_handles_labels()

        size_ax.legend(handle, label, labelspacing=1.2, title=self.size_title, borderpad=1,
                       frameon=True, framealpha=0.6, edgecolor="k", facecolor="w", loc='right')

    def _plot_colorbar(self, cbar_ax: Axes,
                       normalize: mcolors.Normalize) -> None:
        """Plots a horizontal colorbar with normalize values"""
        cmap = plt.get_cmap(self.cmap)

        mappable = ScalarMappable(norm=normalize, cmap=cmap)

        matplotlib.colorbar.Colorbar(
            cbar_ax,
            mappable=mappable,
            orientation="horizontal"
        )

        cbar_ax.set_title(self.colorbar_title, fontsize="small")

        cbar_ax.xaxis.set_tick_params(labelsize="small")



"""
Deprecation
=============================

This module provides decorators for marking classes and functions as deprecated.
It also offers a decorator to support deprecated argument names in functions by
automatically mapping them to their new names and issuing a warning.

Functions and Decorators
------------------------

Class Example:

.. code-block:: python

    @deprecated_class(new="NewClass")
    class OldClass:
        pass

Function Example:

.. code-block:: python

    @deprecated_func(new="new_function")
    def old_function():
        pass

Aliases Example:

    .. code-block:: python

        @deprecated_aliases(old_arg="new_arg")
        def some_function(new_arg):
            pass

        # then you can use old args but deprecate warning
        some_function(old_arg=...)


"""
import functools
import inspect
import warnings

__all__ = ['deprecated_class',
           'deprecated_func',
           'deprecated_aliases']


def deprecated_class(*, new: str | None = None,
                     remarks: str | None = None,
                     removal_version: str | None = None):
    """Mark deprecated class

    :param new: The renamed new usage
    :param remarks: Further remarks to be shown
    :param removal_version: Optional version or date when the function is planned to be removed
    """

    def _decorator(cls):
        ori_init = cls.__init__

        @functools.wraps(ori_init)
        def new_init(self, *args, **kwargs):
            msg = _build_deprecated_message(cls.__qualname__, new, remarks, removal_version)

            warnings.warn(
                msg,
                DeprecationWarning,
                stacklevel=2
            )

            ori_init(self, *args, **kwargs)

        cls.__init__ = new_init

        #
        if cls.__doc__ is None:
            cls.__doc__ = f"DEPRECATED ({remarks}). "
        else:
            cls.__doc__ = f"DEPRECATED ({remarks}). " + cls.__doc__

        return cls

    return _decorator


def deprecated_func(*, new: str | None = None,
                    remarks: str | None = None,
                    removal_version: str = None):
    """Mark deprecated functions.

    :param new: The renamed new usage
    :param remarks: The reason why the function is deprecated
    :param removal_version: Optional version or date when the function is planned to be removed
    """

    def _decorator(f):

        @functools.wraps(f)
        def _deprecated_func(*args, **kwargs):
            msg = _build_deprecated_message(f.__qualname__, new, remarks, removal_version)

            warnings.warn(
                msg,
                DeprecationWarning,
                stacklevel=2
            )

            return f(*args, **kwargs)

        if f.__doc__ is None:
            _deprecated_func.__doc__ = f"DEPRECATED ({remarks}). "
        else:
            _deprecated_func.__doc__ = f"DEPRECATED ({remarks}). " + f.__doc__

        return _deprecated_func

    return _decorator


def _build_deprecated_message(target: str,
                              alternation: str | None = None,
                              remarks: str | None = None,
                              removal: str | None = None) -> str:
    msg = f'{target} is deprecated'

    if removal is not None:
        msg += f' and will be removed in a future release (after version {removal}).'
    else:
        msg += '.'

    if alternation is not None:
        msg += f' Please use "{alternation}" instead.'

    if remarks is not None:
        msg += f' NOTE: {remarks}.'

    return msg


def deprecated_aliases(**aliases: str):
    """
    Mark deprecated argument names and map them to new argument names in a function

    This decorator allows you to support old argument names while transitioning to new ones.
    It will raise a ``DeprecationWarning`` if the old argument name is used and will automatically
    map the value to the new argument name.

    :param aliases: mapping of `old argument names` to `new argument names`.
    :return: Decorated function with support for deprecated argument names.

    :raises RuntimeError: If a new argument name does not exist in the function's signature.
    :raises ValueError: If both old and new argument names are provided simultaneously.
    """

    def _decorator(f):
        sig = inspect.signature(f)

        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            for old_arg, new_arg in aliases.items():

                if new_arg not in sig.parameters:
                    raise RuntimeError(f'New argument: {new_arg} is not in the function arg.')

                if new_arg in kwargs and old_arg in kwargs:
                    raise ValueError(f'Cannot specify both {old_arg} and {new_arg} at the same time')

                if old_arg in kwargs:
                    warnings.warn(
                        f'"{old_arg}" is deprecated and will be removed in future version. Use "{new_arg}" instead',
                        DeprecationWarning,
                        stacklevel=2
                    )

                    # If the new argument is not already in kwargs, move the value from the old argument
                    if new_arg not in kwargs:
                        kwargs[new_arg] = kwargs.pop(old_arg)

            return f(*args, **kwargs)

        return _wrapper

    return _decorator



import abc
from pathlib import Path
from typing import Iterable, Literal

import cv2
import numpy as np

from argclz import AbstractParser, argument

__all__ = ['AbstractSegmentationOptions']


class AbstractSegmentationOptions(AbstractParser, metaclass=abc.ABCMeta):
    DESCRIPTION = 'Base Cellular Segmentation Option'

    # ---- GROUP_IO -----------------------------
    GROUP_IO = 'Data I/O Options'
    EX_GROUP_SOURCE = 'EX_GROUP_SOURCE'

    file: Path = argument(
        '--file', '--image_path',
        ex_group=EX_GROUP_SOURCE,
        group=GROUP_IO,
        help='image file path'
    )

    directory: Path = argument(
        '--dir',
        ex_group=EX_GROUP_SOURCE,
        group=GROUP_IO,
        help='directory for batch imaging processing'
    )

    directory_suffix: Literal['.tif', '.tiff', '.png'] = argument(
        '--dir_suffix',
        default='.tif',
        group=GROUP_IO,
        help='suffix in the directory for batch mode'
    )

    save_ij_roi: bool = argument(
        '--save_rois',
        group=GROUP_IO,
        help='if save also the imageJ/Fiji compatible .roi file'
    )

    # ---- OTHERS -----------------------------
    model: str = argument(
        '--model',
        help='which pretrained model for evaluation'
    )

    invalid_existed_result: bool = argument(
        '--invalid',
        help='force re-evaluate the result'
    )

    no_normalize: bool = argument(
        '--no_norm',
        help='not do percentile-based image normalization'
    )

    napari_view: bool = argument(
        '--napari',
        help='view result by napari GUI, only available in single file mode'
    )

    @property
    def file_mode(self) -> bool:
        """flag file mode"""
        return self.file is not None

    @property
    def batch_mode(self) -> bool:
        """flag batch mode"""
        return self.directory is not None

    @property
    def with_norm(self) -> bool:
        """flag normalize image"""
        return not self.no_normalize

    def process_image(self, to_gray: bool = True) -> np.ndarray:
        """Process the image for segmentation.

        :return: `Array[Any, [H, W]]` or `Array[Any, [H, W, C]]`
        """
        return process_image(
            cv2.imread(str(self.file)),
            to_gray=to_gray,
            norm=self.with_norm
        )

    def foreach_process_image(self, to_gray: bool = True) -> Iterable[tuple[Path, np.ndarray]]:
        """
        Iterates over image files in the specified directory, processes each image, and yields
        the file path along with the processed image. The processing can include grayscale
        conversion and normalization based on the provided parameters.

        :param to_gray: Flag indicating whether the images should be converted to grayscale.
        :return: An iterable of tuples where each tuple includes the file path as a `Path` object
            and the processed image as a numpy array.
        :rtype: Tuple of filepath, image_array (`Array[Any, [H, W]]` or `Array[Any, [H, W, C]]`) generator
        """
        for file in self.directory.glob(f'*{self.directory_suffix}'):
            img = process_image(
                cv2.imread(str(file)),
                to_gray=to_gray,
                norm=self.with_norm
            )
            yield file, img

    @abc.abstractmethod
    def seg_output(self, filepath: Path) -> Path:
        """
        Get segmented output save path

        :param filepath: filepath for image
        :return: segmented output save path
        """
        pass

    def ij_roi_output(self, filepath: Path) -> Path:
        """
        Get imageJ/Fiji ``.roi`` output save path

        :param filepath: filepath for image
        :return: ij roi output save path
        """
        return filepath.with_suffix('.roi')

    @abc.abstractmethod
    def eval(self) -> None:
        """eval the model in single file or batch files in directory, and save the results"""
        pass

    @abc.abstractmethod
    def launch_napari(self, **kwargs):
        """run napari GUI viewer"""
        pass


def process_image(img: np.ndarray,
                  to_gray: bool = True,
                  norm: bool = True) -> np.ndarray:
    """
    Pre process the image for segmentation.

    :param img: image array, `Array[Any, [H, W]]` or `Array[Any, [H, W, C]]`
    :param to_gray: to grayscale. default to True
    :param norm: to min-max normalize, default to True
    :return: pre-processed image array, `Array[Any, [H, W]]` or `Array[Any, [H, W, C]]`
    """
    if to_gray and img.ndim == 3:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    if norm:
        from csbdeep.utils import normalize
        img = normalize(img, clip=True, pmin=20)

    return img




import platform
from contextlib import contextmanager
from pathlib import Path
from typing import Literal, ContextManager

import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray

__all__ = ['plot_figure',
           'ax_set_default_style',
           'ax_merge']

MPL_BACKEND_TYPE = Literal[
    'GTK3Agg', 'GTK3Cairo', 'GTK4Agg', 'GTK4Cairo', 'MacOSX', 'nbAgg',
    'QtAgg', 'QtCairo', 'Qt5Agg', 'Qt5Cairo',
    'TkAgg', 'TkCairo',
    'WebAgg',
    'WX', 'WXAgg', 'WXCairo'
]


@contextmanager
def plot_figure(output: Path | None,
                *args,
                set_square: bool = False,
                set_equal_scale: bool = False,
                win_backend: MPL_BACKEND_TYPE | None = None,
                dpi: int | None = None,
                default_style: bool = True,
                tight_layout: bool = True,
                font_sans_serif: str | None = 'Arial',
                **kwargs) -> ContextManager[Axes] | ContextManager[NDArray[Axes]]:
    """
    Context manager for creating and saving a matplotlib figure

    **Example**

    >>> fig_output = Path('output.png')
    >>> with plot_figure(fig_output) as ax:
    ...     ax.plot([0, 10], [0, 10])

    generate output.png

    :param output: Path to save the output figure. If None, the figure will be shown.
    :param args: Arguments for ``plt.subplots()``
    :param set_square: If True, set the plot to be square
    :param set_equal_scale: If True, set equal scaling for x and y axes
    :param win_backend: Backend to handle backend issues in Windows
        If high resolution image (WXAgg), otherwise keep normal pdf output (WXCario)
    :param dpi: DPI for saving the figure
    :param default_style: If True, apply default style to the axes
    :param tight_layout: If True, apply tight layout to the figure
    :param font_sans_serif: Font style. If None, then use default from mplrc
    :param kwargs: Additional keyword arguments for ``plt.subplots()``
    :return: Single or array of matplotlib Axes object
    """
    if win_backend is not None and platform.system() == 'Windows':
        matplotlib.use(win_backend)

    if font_sans_serif is not None:
        plt.rcParams['font.sans-serif'] = font_sans_serif

    fig, ax = plt.subplots(*args, **kwargs)

    if default_style:
        if isinstance(ax, np.ndarray):
            for _ax in ax.ravel():
                ax_set_default_style(_ax)
        else:
            ax_set_default_style(ax)

    #
    try:
        yield ax
    except Exception as e:
        raise RuntimeError(f'An error occurred while plotting {e}')
    else:

        if set_square:
            if isinstance(ax, np.ndarray):
                for _ax in ax.ravel():
                    _set_square(_ax)
            else:
                _set_square(ax)

        if set_equal_scale:
            if isinstance(ax, np.ndarray):
                for _ax in ax.ravel():
                    _ax.set_aspect('equal')
            else:
                ax.set_aspect('equal')

        # io
        if output is None:
            if tight_layout:
                plt.tight_layout()
            plt.show()

        else:
            while True:
                if tight_layout:
                    plt.tight_layout()

                # for batch calling pulse
                try:
                    plt.savefig(output, dpi=dpi)
                    break
                except OSError as e:
                    input(f'{repr(e)}! press to continue')

    finally:
        plt.clf()
        plt.close('all')


def _set_square(ax: Axes) -> None:
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')


def ax_set_default_style(ax: Axes):
    """
    Default `Axes` style

    :param ax: ``Axes``
    """
    if 'polar' in ax.spines.keys():
        pass
    else:
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.xaxis.set_tick_params(width=1)
        ax.yaxis.set_tick_params(width=1)
        for axis in ['bottom', 'left']:
            ax.spines[axis].set_linewidth(1)


# ========= #
# AxesMerge #
# ========= #


class AxesMergeHelper:
    def __init__(self, ax: np.ndarray):
        """
        :param ax: `Array[Axes, G]`
        """
        self.__fig = ax.ravel()[0].figure
        self.__gs = ax.ravel()[0].get_gridspec()
        self.__ax = ax

    def __getitem__(self, item) -> Axes:
        for ax in self.__ax[item].ravel():
            ax.set_visible(False)

        ret = self.__fig.add_subplot(self.__gs[item])
        ax_set_default_style(ret)
        return ret


def ax_merge(ax: np.ndarray) -> AxesMergeHelper:
    """
    Subplots ``Axes`` merge

    `Dimension parameters`:

        R = number of rows

        C = number of columns

        G = R * C


    **Example of 5x3 grid subplots**

    +-----------------+-----------------+-----------------+
    |    Subplot 1    |    Subplot 1    |    Subplot 4    |
    |    ax1 (3x2)    |    ax1 (3x2)    |    ax4 (5x1)    |
    +-----------------+-----------------+-----------------+
    |    Subplot 1    |    Subplot 1    |    Subplot 4    |
    |    ax1 (3x2)    |    ax1 (3x2)    |    ax4 (5x1)    |
    +-----------------+-----------------+-----------------+
    |    Subplot 1    |    Subplot 1    |    Subplot 4    |
    |    ax1 (3x2)    |    ax1 (3x2)    |    ax4 (5x1)    |
    +-----------------+-----------------+-----------------+
    |    Subplot 2    |    Subplot 3    |    Subplot 4    |
    |    ax2 (2x1)    |    ax3 (2x1)    |    ax4 (5x1)    |
    +-----------------+-----------------+-----------------+
    |    Subplot 2    |    Subplot 3    |    Subplot 4    |
    |    ax2 (2x1)    |    ax3 (2x1)    |    ax4 (5x1)    |
    +-----------------+-----------------+-----------------+


    >>> with plot_figure(None, 5, 3) as _ax:
    ...     ax1 = ax_merge(_ax)[:3, :2]
    ...     ax1.plot(np.arange(5), 'ro')
    ...     ax1.set_title('subplot 1')
    ...
    ...     ax2 = ax_merge(_ax)[3:, 1]
    ...     ax2.imshow(np.random.sample((10, 10)))
    ...     ax2.set_title('subplot 2')
    ...
    ...     ax3 = ax_merge(_ax)[3:, 2]
    ...     ax3.plot(np.arange(10), 'g*')
    ...     ax3.set_title('subplot 3')
    ...
    ...     D = np.random.normal((3, 5, 4), (0.75, 1.00, 0.75), (200, 3))
    ...     ax4 = ax_merge(_ax)[:, -1]
    ...     ax4.violinplot(D, [2, 4, 6], widths=2, showmeans=False, showmedians=False, showextrema=False)
    ...     ax4.set_title('subplot 4')


    :param ax: `Array[Axes, G]`
    :return: ``AxesMergeHelper``
    """
    return AxesMergeHelper(ax)



from __future__ import annotations

from typing import NamedTuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scipy.interpolate import interp1d
from typing_extensions import Self

from neuralib.model.rastermap import RasterMapResult
from neuralib.plot import plot_figure
from neuralib.typing import PathLike

__all__ = [
    'plot_rastermap',
    'plot_cellular_spatial',
    'plot_wfield_spatial',
    'RasterMapPlot',
    'BehavioralVT'
]


def plot_rastermap(result: RasterMapResult,
                   act_time: np.ndarray, *,
                   time_range: tuple[float, float] | None = None,
                   behaviors: list[BehavioralVT] | None = None,
                   output: PathLike | None = None):
    """
    plot the rastermap result with behavioral measurements

    :param result: :class:`~.core.RasterMapResult`
    :param act_time: neural activity time array. should be the same T as neural_activity when run the rastermap
    :param time_range: time range for plotting (START,END)
    :param behaviors: list of :class:`~BehavioralVT`
    :param output: output path for figure save. If None then show
    """
    plotter = RasterMapPlot(result, act_time, time_range, behaviors, output)
    plotter.plot_rastermap()


def plot_cellular_spatial(result: RasterMapResult,
                          xpos: np.ndarray,
                          ypos: np.ndarray,
                          ax: Axes | None = None,
                          output: PathLike | None = None,
                          **kwargs):
    """
    Plot spatial location of each cell cluster by rastermap

    :param result: :class:`~.core.RasterMapResult`
    :param xpos: soma central X position.`Array[float, N]`
    :param ypos: soma central Y position.`Array[float, N]`
    :param ax: ``Axes``
    :param output: output path for figure save. If None then show
    :param kwargs: additional arguments pass to ``ax.set()``
    :return:
    """
    if ax is None:
        _, ax = plt.subplots()

    ax.scatter(xpos, ypos, s=8, c=result.embedding, cmap="gist_ncar", alpha=0.25)
    ax.invert_yaxis()
    ax.set(**kwargs)
    ax.set_aspect('equal')

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


def plot_wfield_spatial(result: RasterMapResult,
                        width: int,
                        height: int,
                        ax: Axes | None = None,
                        output: PathLike | None = None,
                        **kwargs):
    """
    Plot spatial location of each pixel cluster by rastermap

    :param result: :class:`~.core.RasterMapResult`
    :param width: sequence image width
    :param height: sequence image height
    :param ax: ``Axes``
    :param output: output path for figure save. If None then show
    :param kwargs: additional arguments pass to ``ax.set()``
    """

    if ax is None:
        _, ax = plt.subplots()

    x = np.arange(width)
    y = np.arange(height)
    xpos, ypos = np.meshgrid(x, y)  # Array[float, [W, H]]

    ax.scatter(xpos, ypos, s=1, c=result.embedding, cmap="gist_ncar", alpha=0.25)
    ax.invert_yaxis()
    ax.set(**kwargs)
    ax.set_aspect('equal')

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


class RasterMapPlot:
    """Plot the rastermap result with behavioral measurements"""

    def __init__(self, result: RasterMapResult,
                 act_time: np.ndarray,
                 time_range: tuple[float, float] | None = None,
                 behaviors: list[BehavioralVT] | None = None,
                 output: PathLike | None = None):
        """
        :param result: class:`~.core.RasterMapResult`
        :param act_time: neural activity time array. should be the same T as neural_activity when run the rastermap
        :param time_range: time range for plotting (START,END)
        :param behaviors: list of :class:`~BehavioralVT`
        """
        self.raster = result
        self.behaviors = behaviors

        self.time_range = time_range or (act_time[0], act_time[-1])

        if time_range is not None:
            self.act_mask = np.logical_and(time_range[0] < act_time, act_time < time_range[1])
        else:
            self.act_mask = np.ones_like(act_time, dtype=np.bool_)

        self.act_time = act_time[self.act_mask]

        self.output = output

    @property
    def super_neurons(self) -> np.ndarray:
        """rastermap sorted 2D array. `Array[float, [N, T]]`"""
        return self.raster.super_neurons[:, self.act_mask]

    def process_behavior(self) -> list[BehavioralVT]:
        """process behavioral measurements, select time range and do the interpolation same shape as neural activity"""
        return [
            it.masking_time(self.time_range).interp_activity(self.act_time)
            for it in self.behaviors
        ]

    def plot_rastermap(self):
        if self.behaviors is not None:
            behavior_list = self.process_behavior()
            n_behaviors = len(behavior_list)
        else:
            behavior_list = None
            n_behaviors = 1

        height_ratios = [1] * n_behaviors + [7]

        with plot_figure(self.output, n_behaviors + 1, 1, gridspec_kw={'height_ratios': height_ratios},
                         tight_layout=False, sharex=True) as _ax:

            if self.behaviors is not None:
                for i, it in enumerate(behavior_list):
                    ax = _ax[i]
                    ax.plot(it.time, it.value, color='k')
                    ax.set_xlim(self.time_range)
                    ax.axis('off')
                    ax.set_title(it.name)
            else:
                _ax[0].axis('off')

            ax = _ax[n_behaviors]
            ax.imshow(
                self.super_neurons,
                cmap='gray_r',
                vmin=0,
                vmax=0.8,
                aspect='auto',
                extent=(self.time_range[0], self.time_range[1], self.raster.n_clusters, 0),
            )
            ax.set(xlabel='time', ylabel='rastermap clusters')

            # colormap
            n_clusters = self.raster.n_clusters
            cluster_colors = plt.get_cmap("gist_ncar", n_clusters)
            cb_ax = inset_axes(ax, width="2%", height="100%", loc="right",
                               bbox_to_anchor=(0.05, 0., 1, 1), bbox_transform=ax.transAxes, borderpad=0)

            cb_ax.imshow(
                np.arange(n_clusters)[:, np.newaxis],
                cmap=cluster_colors,
                aspect="auto"
            )
            cb_ax.axis("off")


class BehavioralVT(NamedTuple):
    name: str
    """name of the behavioral variable"""
    time: np.ndarray
    """time array. `Array[float, T]`"""
    value: np.ndarray
    """value array. `Array[float, T]`"""

    def masking_time(self, t: tuple[float, float]) -> Self:
        """
        mask given time range
        :param t: (START,END) time range
        :return:
        """
        mx = np.logical_and(t[0] < self.time, self.time < t[1])
        return self._replace(time=self.time[mx], value=self.value[mx])

    def interp_activity(self, act_time: np.ndarray) -> Self:
        """
        interpolation to another activity array. i.e., neural activity
        :param act_time: activity array. `Array[float, T']`
        :return:
        """
        v = interp1d(self.time, self.value, bounds_error=False, fill_value=0)(act_time)
        return self._replace(time=act_time, value=v)



from argclz.commands import parse_command_args
from neuralib.imaging.widefield.align import NapariAlignmentOptions

from .fft_view import WideFieldFFTViewOption


def main():
    parsers = dict(
        align=NapariAlignmentOptions,
        fft=WideFieldFFTViewOption
    )

    parse_command_args(
        parsers=parsers,
        description='widefield tools',
        usage="""
        Usage Examples:

        View HSV map in FFT:
        $ neuralib_widefield fft <FILE>
        
        Alignment napari for image sequence
        $ neuralib_widefield align <FILE>
        
        """
    )


if __name__ == '__main__':
    main()



import numpy as np

__all__ = ['SequenceFFT']


class SequenceFFT:

    def __init__(self, seq: np.ndarray,
                 component: int = 1):
        """
        :param seq: Image Sequence. `Array[float, [F, H, W]]`
        :param component: Which Fourier component (frequency) to use, defaults to 1 (primary oscillatory component)
        """
        self.seq = seq
        self._component = component

        # cache
        self._freq_map: np.ndarray | None = None
        self._intensity: np.ndarray | None = None

    @property
    def component(self) -> int:
        return self._component

    @property
    def n_frames(self) -> int:
        return self.seq.shape[0]

    @property
    def height(self) -> int:
        return self.seq.shape[1]

    @property
    def width(self) -> int:
        return self.seq.shape[2]

    def get_freq_map(self) -> np.ndarray:
        """
        Takes in trial-averaged sequence and calculates the Fourier Transform,
        basically transforming each pixel’s intensity over time (across frames)

        :return: `Array[float, [H, W]]`
        """
        if self._freq_map is None:
            fft_movie = np.fft.fft(self.seq, axis=0)
            freq = fft_movie[self.component]
            self._freq_map = freq
        return self._freq_map

    def get_intensity(self) -> np.ndarray:
        """Computes the magnitude of the frequency component.
        i.e., Strength of the responsiveness (frame-wise averaged)"""
        if self._intensity is None:
            self._intensity = np.abs(self.get_freq_map()) / self.n_frames
        return self._intensity

    def get_phase(self) -> np.ndarray:
        """Computes the phase of the selected frequency component.
        For example, the spatial locations of the visual field.

        :return: `Array[float, [H, W]]`
        """
        return -1 * np.angle(self.get_freq_map()) % (2 * np.pi)

    def as_colormap(self, *,
                    saturation_factor: float = 0.3,
                    value_perc: float = 98,
                    saturation_perc: float = 90,
                    to_rgb: bool = True) -> np.ndarray:
        """
        Generates an HSV colormap representation of the Fourier transform results.

        - Hue (H) represents the phase of the oscillation.
        - Saturation (S) represents the intensity raised to a power (`saturation_factor`).
        - Value (V) represents the normalized intensity of the frequency component.

        The generated colormap provides a visualization of frequency components
        across spatial locations.

        :param saturation_factor: Exponent applied to the intensity for controlling saturation scaling
        :param value_perc: Percentile threshold to normalize intensity values
        :param saturation_perc: Percentile threshold to normalize saturation values
        :param to_rgb: If True, converts the HSV colormap to RGB using OpenCV
        :return:
        """
        intensity = self.get_intensity()

        h = self.get_phase() / (2 * np.pi)  # mapping value from [0, 2pi] to [0, 1]
        s = intensity ** saturation_factor
        v = intensity

        # small value to black / white
        v /= np.percentile(intensity, value_perc)
        s /= np.percentile(s, saturation_perc)

        color_map = np.stack([h, s, v], axis=2).astype(np.float32)

        if to_rgb:
            import cv2
            color_map = np.clip(color_map, 0, 1) * 255
            color_map = cv2.cvtColor(color_map.astype(np.uint8), cv2.COLOR_HSV2RGB_FULL)

        return color_map



import numpy as np
import scipy
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

__all__ = [
    'peri_onset_1d',
    'plot_peri_onset_1d'
]


def plot_peri_onset_1d(event_time: np.ndarray,
                       act_time: np.ndarray,
                       act: np.ndarray, *,
                       pre: float = 3,
                       post: float = 5,
                       plot_all: bool = False,
                       bins: int = 100,
                       with_fill_between: bool = True,
                       ax: Axes | None = None,
                       **kwargs) -> tuple[np.ndarray, np.ndarray]:
    """
    Plot peri-event 1D activity

    :param event_time: peri-event time array. `Array[float, P]`
    :param act_time: activity time array. `Array[float, T]`
    :param act: activity array. `Array[float, T]`
    :param pre: peri-event time before
    :param post: peri-event time after
    :param plot_all: plot velocity every laps
    :param bins:  number of bins per trial in a given time bins (peri-left + peri-right)
    :param with_fill_between: fill_between for the sem
    :param ax: ``Axes``
    :param kwargs: additional arguments to ``ax.plot()``
    :return: x (`Array[float, B]`) and average velocity (`Array[float, B]`)
    """
    x = np.linspace(-pre, post, bins)
    v = peri_onset_1d(event_time, act_time, act, bins, pre, post)
    v_mean = np.nanmean(v, axis=0)

    if ax is None:
        fig, ax = plt.subplots()

    ax.plot(x, v_mean, **kwargs)
    ax.axvline(0, color="r", linestyle="--", zorder=1)

    if plot_all:
        for vlap in v:
            ax.plot(x, vlap, color='grey', alpha=0.1)

    if with_fill_between:
        v_sem = scipy.stats.sem(v)
        ax.fill_between(x, v_mean + v_sem, v_mean - v_sem, alpha=0.3)

    return x, v_mean


def peri_onset_1d(event_time: np.ndarray,
                  act_time: np.ndarray,
                  act: np.ndarray,
                  bins: int = 100,
                  pre: float = 3,
                  post: float = 5) -> np.ndarray:
    """
    Get Peri-event 1D activity

    :param event_time: peri-event time array. `Array[float, P]`
    :param act_time: activity time array. `Array[float, T]`
    :param act: activity array. `Array[float, T]`
    :param bins: number of bins per trial in a given time bins (peri-left + peri-right)
    :param pre: peri-event time before
    :param post: peri-event time after
    :return: peri-reward activity. `Array[float, [P, B]]`. B = number of bins per trial in a given time bins (peri-left + peri-right)

    """
    ret = np.zeros((len(event_time), bins))
    for i, time in enumerate(event_time):
        left = time - pre
        right = time + post
        time_mask = np.logical_and(left < act_time, act_time < right)
        t = act_time[time_mask]
        v = act[time_mask]

        hist, edg = np.histogram(t, bins, range=(left, right), weights=v)
        occ = np.histogram(t, edg)[0]
        hist /= occ
        hist[np.isnan(hist)] = 0
        ret[i] = hist

    return ret



from pathlib import Path

import cv2
import numpy as np
from PIL import Image
from tifffile import tifffile
from tqdm import tqdm

from neuralib.imglib.norm import normalize_sequences
from neuralib.typing import PathLike

__all__ = [
    'load_sequence',
    'read_avi',
    'read_pdf',
    'write_avi',
    'tif_to_gif',
    'gif_show'
]


def load_sequence(path: PathLike, suffix: str = '*.tif') -> np.ndarray:
    """
    Recursively load a TIFF sequence from a file or directory using memory mapping.

    - If `path` is a TIFF file, returns a memmapped NumPy array.
    - If `path` is a directory, loads and stacks all valid TIFF files matching `suffix`.
    - Ignores hidden files (starting with '.') or AppleDouble files (._).

    :param path: Path to a .tif file or a directory of .tif files.
    :param suffix: Glob pattern for selecting TIFF files (default: '*.tif').
    :return: Stacked array of TIFF frames. `Array[Any, [H, W]]`
    :raise RuntimeError: If the path is not found or no valid TIFFs were found.
    """
    path = Path(path)

    if path.is_file():
        return tifffile.memmap(path)

    elif path.is_dir():
        files = sorted([
            f
            for f in path.glob(suffix)
            if f.is_file() and not f.name.startswith('._') and not f.name.startswith('.')
        ])

        if not files:
            raise FileNotFoundError(f'No valid TIFF files found in directory: {path}')

        arrays = []
        for f in tqdm(files, desc='Loading TIFF sequence', unit='file'):
            try:
                arrays.append(tifffile.memmap(f))
            except Exception as e:
                print(f'Skipped invalid TIFF file: {f.name} ({e})')

        if not arrays:
            raise ValueError(f'All files in {path} failed to load as TIFFs.')

        return np.vstack(arrays)

    else:
        raise FileNotFoundError(f'Path not found: {path}')


def read_avi(avi_file: PathLike, grey_scale: bool = True) -> np.ndarray:
    """Read a sequences file (i.e., AVI/MP4) into an array, converting to grayscale if specified.

    :param avi_file: Path to the AVI file.
    :param grey_scale: Whether to convert frames to grayscale.
    :return: Sequences array. `Array[uint8, [F, W, H]|[F, W, H, 3]]`
    """
    cap = cv2.VideoCapture(str(avi_file))
    if not cap.isOpened():
        raise RuntimeError(f'Error opening AVI file: {avi_file}')

    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    if grey_scale:
        shape = (frame_count, height, width)
    else:
        shape = (frame_count, height, width, 3)

    frames = np.empty(shape, dtype=np.uint8)
    i = 0
    try:
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            if grey_scale:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            frames[i] = frame
            i += 1
    finally:
        cap.release()

    return frames


def read_pdf(file: PathLike,
             single_image: bool = True,
             poppler_path: str | None = None,
             **kwargs) -> np.ndarray:
    """
    Read pdf as an image array

    :param file:
    :param single_image:
    :param poppler_path: if poppler not installed in a system level. Specify the path.
        for example: ``Release-23.08.0-0\poppler-23.08.0\Library\bin``

        .. seealso::

            `<https://stackoverflow.com/questions/53481088/poppler-in-path-for-pdf2image>`_
    :param kwargs: pass through `convert_from_path`
    :return: Image array
    """
    from pdf2image import convert_from_path
    from pdf2image.exceptions import PDFInfoNotInstalledError

    if single_image:
        try:
            conv = convert_from_path(file, **kwargs)[0]
        except PDFInfoNotInstalledError:
            if poppler_path is None:
                raise RuntimeError('download poppler first. or using apt-get / brew depending on OS')

            conv = convert_from_path(file, poppler_path=poppler_path, **kwargs)[0]
    else:
        raise NotImplementedError('multi-pages pdf not currently support')

    return cv2.cvtColor(np.array(conv), cv2.COLOR_BGR2RGB)


def tif_to_gif(image_file: PathLike,
               output_path: PathLike,
               fps: int = 30,
               **kwargs) -> None:
    """Convert tif sequences to GIF"""
    import imageio

    frames = imageio.mimread(image_file, memtest=False)
    frames = normalize_sequences(frames, **kwargs)
    imageio.mimsave(output_path, frames, duration=1 / fps)


def gif_show(file: PathLike) -> None:
    """
    Display GIF file in a cv2 window

    :param file: Path to the GIF file to be displayed.
    :raises ValueError: If the provided file is not a GIF.
    """
    if isinstance(file, str):
        file = Path(file)

    if not file.suffix == '.gif':
        raise ValueError('must be gif file!')

    gif = Image.open(file)
    while True:
        try:
            frame = np.array(gif.convert('RGB'))
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
            cv2.imshow('GIF', frame)

            if cv2.waitKey(100) & 0xFF == ord('q'):
                break

            gif.seek(gif.tell() + 1)

        except EOFError:
            gif.seek(0)

    gif.close()
    cv2.destroyAllWindows()


def write_avi(file_path: str, frames: np.ndarray, fps: int = 30.0) -> None:
    """
    Write a sequence of frames to an AVI file.

    :param file_path: The path where the AVI file will be saved.
    :param frames: `Array[uint, [F, W, H]|[F, W, H, 3]]` containing the frames.
    :param fps: The frames per second (frame rate) of the output video.
    """
    if not file_path.endswith('avi'):
        raise ValueError('file path need to write to *.avi')

    if not len(frames):
        raise ValueError("Frames array is empty.")

    height, width = frames[0].shape[:2]

    if frames[0].ndim == 3:
        is_color = True
    else:
        is_color = False

    fourcc = cv2.VideoWriter_fourcc(*'MJPG')
    out = cv2.VideoWriter(file_path, fourcc, fps, (width, height), isColor=is_color)

    for frame in tqdm(frames, desc='write frames to avi', unit='frames'):
        # frame = frame[:, :, [2, 1, 0]]  # Convert RGB to BGR
        out.write(frame)

    out.release()



from __future__ import annotations

import datetime
from pathlib import Path
from typing import Literal, TypedDict, final

import attrs
import numpy as np
import polars as pl
from typing_extensions import Self

from neuralib.imaging.registration import CellularCoordinates, get_cellular_coordinate
from neuralib.typing import PathLike
from neuralib.util.verbose import fprint

__all__ = [
    'SIGNAL_TYPE',
    'CALCIUM_TYPE',
    'read_suite2p',
    #
    'Suite2pGUIOptions',
    'Suite2pRoiStat',
    'Suite2PResult',
    #
    'get_s2p_coords'
]

SIGNAL_TYPE = Literal['df_f', 'spks']
"""imaging fluorescence signal type"""

CALCIUM_TYPE = Literal['baseline', 'transient']


def read_suite2p(directory: PathLike, *,
                 cell_prob_thres: float | None = 0.5,
                 red_cell_threshold: float = 0.65,
                 channel: int = 0,
                 runtime_check_frame_rate: float | None = 30.0) -> Suite2PResult:
    """
    Load suite2p result from directory

    :param directory: Directory contain all the s2p output files. e.g., \*/suite2p/plane[P]
    :param cell_prob_thres: Cell probability. If float type, mask for the value in ``iscell[:, 1]``.
                If None, use the binary criteria in GUI output
    :param red_cell_threshold: Red cell threshold
    :param channel: channel (PMT) Number for the functional channel. i.e., 0 if GCaMP, 1 if jRGECO in scanbox setting
    :param runtime_check_frame_rate: if not None, check frame rate lower-bound to make sure the s2p runconfig
    :return: :class:`Suite2PResult`
    """
    return Suite2PResult.load(directory, cell_prob_thres, red_cell_threshold, channel, runtime_check_frame_rate)


@final
@attrs.frozen
class Suite2PResult:
    """suite2p result container

    `Dimension parameters`:

        N: number of neurons

        F: number pf frames

        W: image width

        H: image height
    """
    directory: Path
    """Directory contain all the s2p output files"""

    f_raw: np.ndarray
    """Fluorescence traces 2D array. `Array[float, [N, F]]`"""

    f_neu: np.ndarray
    """Neuropil fluorescence traces 2D array. `Array[float, [N, F]]`"""

    spks: np.ndarray
    """Deconvolved activity 2D array. `Array[float, [N, F]]`"""

    stat: np.ndarray
    """GUI imaging after registration, i.e., x, ypixel., etc. `Array[Suite2pRoiStat, N]`"""

    ops: Suite2pGUIOptions
    """GUI options"""

    iscell: np.ndarray
    """Cell probability for each ROI. `Array[float, [N, 2]]`"""

    cell_prob_thres: float | None
    """Cell probability threshold for loading the data"""

    redcell: np.ndarray | None = attrs.field(default=None)
    """Red cell probability 2D array. `Array[float, [N, 2]]`"""

    redcell_threshold: float | None = attrs.field(default=None)
    """Red cell probability threshold"""

    runtime_frate_check: float | None = attrs.field(default=None)
    """If not None, check frame rate lower bound"""

    def __attrs_post_init__(self):
        if self.runtime_frate_check is not None:
            self._check_frame_rate()

    def _check_frame_rate(self):
        """User specific check suite2p config is correct"""
        if self.fs * self.n_plane > self.runtime_frate_check:
            fprint(f'the fr: {self.fs} and n_etl: {self.n_plane} might not set properly in suite2p,'
                   f'check output ops.json', vtype='error')
            raise RuntimeError('fs and n_plane are not set properly in suite2p')

    @classmethod
    def load(cls, directory: PathLike,
             cell_prob_thres: float | None = 0.5,
             red_cell_threshold: float = 0.65,
             channel: int = 0,
             runtime_check_frame_rate: float | None = 30.0) -> Self:
        """
        Load suite2p result from directory

        :param directory: Directory contain all the s2p output files. e.g., \*/suite2p/plane[P]
        :param cell_prob_thres: Cell probability. If float type, mask for the value in ``iscell[:, 1]``.
                    If None, use the binary criteria in GUI output
        :param red_cell_threshold: Red cell threshold
        :param channel: channel (PMT) Number for the functional channel. i.e., 0 if GCaMP, 1 if jRGECO in scanbox setting
        :param runtime_check_frame_rate: if not None, check frame rate lower-bound to make sure the s2p runconfig
        :return: :class:`Suite2PResult`
        """
        if not isinstance(directory, Path):
            directory = Path(directory)

        if channel == 0:
            F = np.load(directory / 'F.npy', allow_pickle=True)
            FNeu = np.load(directory / 'Fneu.npy', allow_pickle=True)
            spks = np.load(directory / 'spks.npy')
            stat = np.load(directory / 'stat.npy', allow_pickle=True)
            ops = np.load(directory / 'ops.npy', allow_pickle=True).tolist()
            iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
            r = directory / 'redcell.npy'
            if r.exists():
                redcell = np.load(r, allow_pickle=True)
            else:
                redcell = None

        elif channel == 1:  # second channel
            F = np.load(directory / 'F_chan2.npy', allow_pickle=True)
            FNeu = np.load(directory / 'Fneu_chan2.npy', allow_pickle=True)
            spks = np.load(directory / 'spks.npy')
            stat = np.load(directory / 'stat.npy', allow_pickle=True)
            ops = np.load(directory / 'ops.npy', allow_pickle=True).tolist()
            iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
            redcell = None
        else:
            raise IndexError(f'{channel} unknown')

        #
        if cell_prob_thres is None:
            x = iscell[:, 0] == 1
        elif isinstance(cell_prob_thres, float):
            x = iscell[:, 1] >= cell_prob_thres
        else:
            raise TypeError(f'invalid type: {type(cell_prob_thres)}')

        F = F[x]
        FNeu = FNeu[x]
        spks = spks[x]
        stat = stat[x]
        iscell = iscell[x]
        if redcell is not None:
            redcell = redcell[x]

        return Suite2PResult(
            directory,
            F,
            FNeu,
            spks,
            stat,
            ops,
            iscell,
            cell_prob_thres,
            redcell,
            red_cell_threshold if channel == 1 else None,
            runtime_check_frame_rate
        )

    @property
    def has_chan2(self) -> bool:
        """if has a second channel"""
        if self.ops['nchannels'] == 2:
            return True
        return False

    @property
    def n_neurons(self) -> int:
        """number of neurons after load.
        could be less than GUI ROI number if use higher cell_prob in
        :meth:`~neuralib.imaging.suite2p.core.Suite2PResult.load()`"""
        return self.f_raw.shape[0]

    @property
    def n_frame(self) -> int:
        """number of frame number"""
        return self.f_raw.shape[1]

    @property
    def cell_prob(self) -> np.ndarray:
        """probability that the ROI is a cell based on the default classifier. `Array[float, N]`"""
        return self.iscell[:, 1]

    @property
    def n_red_neuron(self) -> int:
        """number of identified neuron based on red cell threshold"""
        if self.has_chan2:
            return np.count_nonzero(self.red_cell_prob >= self.redcell_threshold)
        else:
            raise RuntimeError('no channel 2')

    @property
    def red_cell_prob(self) -> np.ndarray | None:
        """red cell probability, `Array[float, N]`"""
        if self.redcell is None:
            return None
        return self.redcell[:, 1]

    @property
    def signal_baseline(self) -> float:
        """Gaussian filter width in seconds"""
        return self.ops['sig_baseline']

    @property
    def window_baseline(self) -> float:
        """window for max/min filter in seconds"""
        return self.ops['win_baseline']

    @property
    def fs(self) -> float:
        """suite2p approximate frame rate per plane, exact value should be checked in .sbx or .mat"""
        return self.ops['fs']

    @property
    def neucoeff(self) -> float:
        """neuropil coefficient, normally should be ~0.7"""
        return self.ops['neucoeff']

    @property
    def prctile_baseline(self) -> float:
        """percentile of trace to use as baseline if ops['baseline'] = constant_percentile"""
        return self.ops['prctile_baseline']

    @property
    def n_plane(self) -> int:
        """number of optical plane"""
        return self.ops['nplanes']

    @property
    def image_width(self) -> int:
        """image width (in pixel)"""
        return self.ops['Lx']

    @property
    def image_height(self) -> int:
        """image height (in pixel)"""
        return self.ops['Ly']

    @property
    def image_mean(self) -> np.ndarray:
        """mean image for chan0(1st). `Array[float, [H, W]]`"""
        return self.ops['meanImg'].T

    @property
    def image_mean_ch2(self) -> np.ndarray:
        """mean image for chan1(2nd). `Array[float, [H, W]]`"""
        return self.ops['meanImg_chan2'].T

    @property
    def indicator_tau(self) -> float:
        """The timescale of the sensor (in seconds)"""
        return self.ops['tau']

    @property
    def rigid_x_offsets(self) -> np.ndarray:
        """x-shifts of recording at each timepoint. `Array[int, F]`"""
        return self.ops['xoff']

    @property
    def rigid_y_offsets(self) -> np.ndarray:
        """y-shifts of recording at each timepoint. `Array[int, F]`"""
        return self.ops['yoff']

    @property
    def rigid_xy_offset(self) -> np.ndarray:
        """peak of phase correlation between frame and reference image at each timepoint. `Array[float, F]`"""
        return self.ops['corrXY']

    @property
    def nonrigid_x_offsets(self) -> np.ndarray:
        """(frames, block_size). `Array[float, F]`"""
        return self.ops['xoff1']

    @property
    def nonrigid_y_offsets(self) -> np.ndarray:
        """`Array[float, F]`"""
        return self.ops['yoff1']

    @property
    def nonrigid_xy_offsets(self) -> np.ndarray:
        """`Array[float, F]`"""
        return self.ops['corrXY1']

    @classmethod
    def load_total_neuron_number(cls,
                                 directory: Path,
                                 cell_prob: float | None = 0.5) -> int:
        """
        Load number of neuron based on iscell.npy

        :param directory: directory contains the iscell.npy
        :param cell_prob: cell probability,
                    bool type: use the binary criteria in GUI output
                    float type: value in ``iscell[:, 1]``
        :return: Number of neurons
        """
        iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
        if cell_prob is None:
            return np.count_nonzero(iscell[:, 0] == 1)
        else:
            return np.count_nonzero(iscell[:, 1] >= cell_prob)

    def get_rois_pixels(self) -> np.ndarray:
        """ROIs pixel (N, 2)"""
        ret = np.zeros((self.n_neurons, 2))
        for i in range(self.n_neurons):
            x = np.mean(self.stat[i]['xpix'])
            y = np.mean(self.stat[i]['ypix'])
            ret[i] = np.array([x, y])

        return ret

    def get_neuron_id_mapping(self) -> pl.DataFrame:
        """
        Retrieves a mapping between neuron IDs and their corresponding raw indices
        based on whether the cell detection probabilities meet a specified threshold.
        If no cell detection probabilities are provided, the mapping assumes all
        indices are valid neurons.

        :return: A Polars DataFrame containing two columns: `neuron_id` and  `raw_index`.
        """
        n = np.arange(len(self.f_raw))
        if self.cell_prob is None:
            return pl.DataFrame([n, n], schema=['neuron_id', 'raw_index'], orient='col')
        else:
            iscell = np.load(self.directory / 'iscell.npy', allow_pickle=True)
            mx = np.nonzero(iscell[:, 1] >= self.cell_prob)[0]
            return pl.DataFrame([n, mx], schema=['neuron_id', 'raw_index'], orient='col')


class Suite2pGUIOptions(TypedDict, total=False):
    """ Suite2p GUI setting.

    .. seealso:: `<https://suite2p.readthedocs.io/en/latest/settings.html>`_"""
    look_one_level_down: float
    fast_disk: str
    delete_bin: bool
    mesoscan: bool
    bruker: bool
    h5py: list
    h5py_key: str
    save_path0: str
    save_folder: str
    subfolders: list
    move_bin: bool
    nplanes: int
    nchannels: int
    functional_chan: int
    tau: float
    fs: float
    force_sktiff: bool
    frames_include: int
    multiplane_parallel: float
    preclassify: float
    save_mat: bool
    save_NWB: float
    combined: float
    aspect: float
    do_bidiphase: bool
    bidiphase: float
    bidi_corrected: bool
    do_registration: int
    two_step_registration: float
    keep_movie_raw: bool
    nimg_init: int
    batch_size: int
    maxregshift: float
    align_by_chan: int
    reg_tif: bool
    reg_tif_chan2: bool
    subpixel: int
    smooth_sigma_time: float
    smooth_sigma: float
    th_badframes: float
    norm_frames: bool
    force_refImg: bool
    pad_fft: bool
    nonrigid: bool
    block_size: tuple[int, int]
    snr_thresh: float
    maxregshiftNR: float
    oneP_reg: bool  # 1Preg
    spatial_hp: int
    spatial_hp_reg: float
    spatial_hp_detect: int
    pre_smooth: float
    spatial_taper: float
    roidetect: bool
    spikedetect: bool
    anatomical_only: float
    sparse_mode: bool
    diameter: float
    spatial_scale: float
    connected: bool
    nbinned: int
    max_iterations: int
    threshold_scaling: float
    max_overlap: float
    high_pass: float
    denoise: bool
    soma_crop: bool
    neuropil_extract: bool
    inner_neuropil_radius: float
    min_neuropil_pixels: int
    lam_percentile: float
    allow_overlap: bool
    use_builtin_classifier: bool
    classifier_path: int
    chan2_thres: float
    baseline: str
    win_baseline: float
    sig_baseline: float
    prctile_baseline: float
    neucoeff: int
    suite2p_version: str
    data_path: list[str]
    sbx_ndeadcols: int
    input_format: str
    save_path: str
    ops_path: str
    reg_file: str
    filelist: list[str]
    nframes_per_folder: np.ndarray
    sbx_ndeadrows: int
    meanImg: np.ndarray
    meanImg_chan2: np.ndarray  # if chan_2
    nframes: int
    Ly: int
    Lx: int
    date_proc: datetime.datetime
    refImg: np.ndarray
    rmin: int
    rmax: int
    yblock: list[np.ndarray]
    xblock: list[np.ndarray]
    nblocks: list[int]
    NRsm: np.ndarray
    yoff: np.ndarray
    xoff: np.ndarray
    corrXY: np.ndarray
    yoff1: np.ndarray
    xoff1: np.ndarray
    corrXY1: np.ndarray
    badframes: np.ndarray
    yrange: list[int]
    xrange: list[int]
    tPC: np.ndarray
    regPC: np.ndarray
    regDX: np.ndarray
    Lyc: int
    Lxc: int
    max_proj: np.ndarray
    Vmax: np.ndarray
    ihop: np.ndarray
    Vsplit: np.ndarray
    Vcorr: np.ndarray
    Vmap: list[np.ndarray]
    spatscale_pix: np.ndarray
    meanImgE: np.ndarray
    timing: dict[str, float]


class Suite2pRoiStat(TypedDict, total=False):
    """Suite2p GUI imaging.

    .. seealso:: `<https://suite2p.readthedocs.io/en/latest/outputs.html#stat-npy-fields>`_
    """
    ypix: np.ndarray
    xpix: np.ndarray
    lam: np.ndarray
    med: list[int, int]
    footprint: float
    mrs: float
    mrs0: float
    compact: float
    solidity: float
    npix: int
    npix_soma: int
    soma_crop: np.ndarray
    overlap: np.ndarray
    radius: float
    aspect_ratio: float
    npix_norm_no_crop: float
    npix_norm: float
    skew: float
    std: float


# def get_s2p_coords(s2p: Suite2PResult,
#                    neuron_list: int | list[int] | slice | np.ndarray | None,
#                    plane_index: int,
#                    factor: float) -> CellularCoordinates:
#     """
#     Get the suite2p coordinates of all cells.
#
#     :param s2p: ``Suite2PResult``
#     :param neuron_list: neuron index or index list/arr. If None, then load all neurons
#     :param plane_index: optic plane index
#     :param factor: pixel to um factor
#     :return: :class:`~neuralib.imaging.cellular_cords.CellularCoordinates`
#     """
#     if neuron_list is None:
#         neuron_list = np.arange(s2p.n_neurons)
#
#     n_neurons = len(neuron_list)
#     xpix = np.zeros(n_neurons)
#     ypix = np.zeros(n_neurons)
#
#     for i, n in enumerate(neuron_list):
#         xpix[i] = np.mean(s2p.stat[i]['xpix'])
#         ypix[i] = np.mean(s2p.stat[i]['ypix'])
#
#     xcord = xpix * factor / 1000  # ap
#     ycord = ypix * factor / 1000  # ml
#
#     if isinstance(neuron_list, int):
#         src_plane_index = plane_index
#     elif isinstance(neuron_list, (list, np.ndarray, slice)):
#         src_plane_index = np.full_like(neuron_list, plane_index)
#     else:
#         raise TypeError('')
#
#     return CellularCoordinates(
#         np.array(neuron_list),
#         xcord,
#         ycord,
#         plane_index=plane_index,
#         unit='mm',
#         source_plane_index=src_plane_index
#     )


def get_s2p_coords(s2p: Suite2PResult,
                   neuron_list: int | list[int] | slice | np.ndarray | None,
                   plane_index: int,
                   factor: float) -> CellularCoordinates:
    """
    Get the suite2p coordinates of all cells.

    :param s2p: ``Suite2PResult``
    :param neuron_list: neuron index or index list/arr. If None, then load all neurons
    :param plane_index: optic plane index
    :param factor: pixel to um factor
    :return: :class:`~neuralib.imaging.cellular_cords.CellularCoordinates`
    """
    if neuron_list is None:
        neuron_list = np.arange(s2p.n_neurons)

    n_neurons = len(neuron_list)
    xpix = np.zeros(n_neurons)
    ypix = np.zeros(n_neurons)

    for i, n in enumerate(neuron_list):
        xpix[i] = np.mean(s2p.stat[i]['xpix'])
        ypix[i] = np.mean(s2p.stat[i]['ypix'])

    xcord = xpix * factor / 1000  # ap
    ycord = ypix * factor / 1000  # ml

    if isinstance(neuron_list, int):
        src_plane_index = plane_index
    elif isinstance(neuron_list, (list, np.ndarray, slice)):
        src_plane_index = np.full_like(neuron_list, plane_index)
    else:
        raise TypeError('')

    return get_cellular_coordinate(
        np.array(neuron_list),
        xcord,
        ycord,
        unit='mm',
        plane_index=src_plane_index
    )



import os

from slack import WebClient
from slack.errors import SlackApiError

from neuralib.typing import PathLike
from neuralib.util.verbose import fprint

__all__ = ['send_slack_message']


def send_slack_message(env_file: PathLike,
                       message: str) -> None:
    """
    Send message to slack channel

    :param env_file: env file with fields ``SLACK_TOKEN`` (i.e., SLACK_TOKEN=xoxb-<USER_SLACK_TOKEN>),
        and ``CHANNEL_ID`` (i.e., ``#general`` or ``ID``)
    :param message: message to send
    :return:
    """

    try:
        from dotenv import load_dotenv
        load_dotenv(dotenv_path=env_file)

        try:
            client = WebClient(token=os.environ['SLACK_TOKEN'])
        except KeyError:
            fprint('To get the slack notification, generate token in local machine first', vtype='warning')
        else:
            try:
                client.chat_postMessage(
                    channel=os.environ['CHANNEL_ID'],
                    text=message,
                )
            except SlackApiError as e:
                print(f"Error sending message: {e.response['error']}")

    except BaseException as e:  # not care, internet disconnected
        fprint(e, vtype='error')



from typing import NamedTuple

import numpy as np
from sklearn.decomposition import TruncatedSVD

__all__ = [
    'SequenceSingularVector',
    'compute_singular_vector'
]


class SequenceSingularVector(NamedTuple):
    """A NamedTuple that represents a singular value decomposition result from image sequences

    `Dimension parameters`:

        W = image width

        H = image height

        F = number of image frame (sequence)

        C = number of components after SVD reduction

    """

    svd: TruncatedSVD
    """Dimensionality reduction using truncated SVD"""
    right_vector: np.ndarray
    """Temporal structure (right singular vectors). `Array[float, [F, C]]`"""
    left_vector: np.ndarray
    """Spatial structure (left singular vectors). `Array[float, [W * H, C]]`"""

    @property
    def singular_value(self) -> np.ndarray:
        """The singular values corresponding to each of the selected components,
        represent the strength/energy of each component. `Array[float, C]`"""
        return self.svd.singular_values_


def compute_singular_vector(sequences: np.ndarray,
                            n_components: int = 128,
                            mean_subtraction: bool = True,
                            **kwargs) -> SequenceSingularVector:
    """
    Performs truncated singular value decomposition (SVD) on a sequence of image frames
    to extract dominant spatial and temporal components.

    :param sequences: A numpy array representing a collection of image frames in a sequence.
                      It has a shape of (n_frames, width, height) where 'n_frames' is the number
                      of frames, and 'width' and 'height' are the dimensions of each frame.
    :param n_components: An integer representing the number of components for Truncated SVD.
                         The default value is 128.
    :param mean_subtraction: A boolean indicating whether to subtract the mean of each frame
    :param kwargs: Keyword arguments passed to ``TruncatedSVD()``.
    :return: A SequenceSingularVector object containing the singular values, the transformed
             components, and the left singular vectors.
    """
    n_frames, width, height = sequences.shape
    seq = sequences.reshape(n_frames, height * width).T

    if mean_subtraction:
        seq = seq - np.mean(seq, axis=1, keepdims=True)

    svd = TruncatedSVD(n_components=n_components, **kwargs)
    Vsv = svd.fit_transform(seq.T)
    U = seq @ (Vsv / (Vsv ** 2).sum(axis=0) ** 0.5)
    U /= (U ** 2).sum(axis=0) ** 0.5

    return SequenceSingularVector(svd, Vsv, U)



import functools
import typing

from .annotation import *
from .table import *
from .table import Table, missing
from .util import resolve_field_type, cast_from_sql, cast_to_sql

__all__ = ['named_tuple_table_class']

T = typing.TypeVar('T')


def named_tuple_table_class(cls):
    """
    A deceorator that deceorate a NamedTuple to be a SQL table.

    Declare a table

    >>> @named_tuple_table_class
    ... class Example(typing.NamedTuple):
    ...     a: typing.Annotated[str, PRIMARY]  # primary key
    ...     b: typing.Annotated[str, UNIQUE]   # unique key
    ...     c: typing.Optional[str]            # nullable key

    """
    ret = NamedTupleTable(cls)
    setattr(cls, '_sql_table', ret)
    return cls


class NamedTupleTable(Table[T], typing.Generic[T]):
    """
    SQL table information for a NamedTuple class that decorated by named_tuple_table_class.
    """

    def __init__(self, table_type: type[T]):
        import typing

        if not hasattr(table_type, '_fields'):
            raise RuntimeError(f'not a NamedTuple {table_type.__name__}')

        self.table_type = table_type
        self._fields: list[Field] = []
        self._unique: list[UniqueConstraint] = []
        self._foreign: list[ForeignConstraint] = []
        self._check: dict[typing.Optional[str], CheckConstraint] = {}

        field_types = typing.get_type_hints(table_type, include_extras=True)
        for i, name in enumerate(getattr(table_type, '_fields')):
            field = self.__setup_column_constraint(table_type, i, name, field_types[name])

            if (constraint := field.get_unique()) is not None:
                self._unique.append(UniqueConstraint(field.name, table_type, [field.name], constraint.conflict))

        self.__setup_table_constraint(table_type)

    def __setup_column_constraint(self, table_type: type[T], i: int, attr_name: str, attr_type) -> Field:
        f_value_missing = missing

        attr_annotations = []
        if typing.get_origin(attr_type) == typing.Annotated:
            attr_annotations = typing.get_args(attr_type)[1:]
            if CURRENT_DATE in attr_annotations:
                f_value_missing = CURRENT_DATE
            elif CURRENT_TIME in attr_annotations:
                f_value_missing = CURRENT_TIME
            elif CURRENT_TIMESTAMP in attr_annotations:
                f_value_missing = CURRENT_TIMESTAMP

        r_type, f_type, not_null = resolve_field_type(attr_type)
        f_value = table_type._field_defaults.get(attr_name, f_value_missing)
        field = Field(table_type, attr_name, r_type, f_type, f_value, not_null, attr_annotations)
        setattr(table_type, attr_name, TableFieldDescriptor(i, field))
        self._fields.append(field)
        return field

    def __setup_table_constraint(self, table_type: type[T]):
        for attr in dir(table_type):
            if callable(attr_value := getattr(table_type, attr)):
                if (constraint := getattr(attr_value, '_sql_foreign', None)) is not None:
                    constraint = make_foreign_constrain(self, attr_value, *constraint)
                    self._foreign.append(constraint)
                    setattr(attr_value, '_sql_foreign', constraint)
                if (constraint := getattr(attr_value, '_sql_check', missing)) is not missing:
                    constraint = make_check_constraint(self, attr_value, *constraint)
                    self._check[constraint.field] = constraint
                    setattr(attr_value, '_sql_check', constraint)
                if (constraint := getattr(attr_value, '_sql_unique', missing)) is not missing:
                    constraint = make_unique_constraint(self, attr_value, *constraint)
                    self._unique.append(constraint)
                    setattr(attr_value, '_sql_unique', constraint)
            elif isinstance(attr_value, property):
                setattr(table_type, attr, setup_table_property(attr_value, table_type))

    @property
    def table_name(self) -> str:
        return self.table_type.__name__

    def table_seq(self, instance: T, fields: list[str] = None) -> tuple[typing.Any, ...]:
        _args = []
        for field, arg in zip(self._fields, instance):
            if field is None or field.name in fields:
                _args.append(cast_to_sql(field.raw_type, field.sql_type, arg))
        return tuple(_args)

    def table_new(self, *args) -> T:
        _args = []
        for field, arg in zip(self._fields, args):
            _args.append(cast_from_sql(field.raw_type, field.sql_type, arg))
        return self.table_type(*_args)

    @property
    def table_fields(self) -> list[Field]:
        return list(self._fields)

    @property
    def table_unique_fields(self) -> list[UniqueConstraint]:
        return list(self._unique)

    @property
    def table_foreign_fields(self) -> list[ForeignConstraint]:
        return list(self._foreign)

    @property
    def table_check_fields(self) -> dict[typing.Optional[str], CheckConstraint]:
        return dict(self._check)


class TableFieldDescriptor:
    __slots__ = '__index', '__field'

    def __init__(self, index: int, field: Field):
        self.__index = index
        self.__field = field

    def __get__(self, instance, owner):
        if instance is None:
            from .expr import SqlField
            return SqlField(self.__field)
        else:
            return instance[self.__index]

    def __str__(self):
        return self.__field.name


def setup_table_property(prop: property, table_type: type) -> property:
    getter = wrap_property_getter(prop.fget, table_type)
    setter = prop.fset
    deleter = prop.fdel
    return property(getter, setter, deleter)


def wrap_property_getter(getter, table_type: type):
    from .expr import SqlExpr

    @functools.wraps(getter)
    def getter_wrapper(self):
        if isinstance(self, type):
            return getter(self)

        ret = getter(self)
        if isinstance(ret, SqlExpr):
            ret = ret.__sql_eval__(self)
        return ret

    getter_wrapper._sql_owner = table_type
    return getter_wrapper



import contextlib
from io import BytesIO
from typing import ContextManager

import requests
from tqdm import tqdm

__all__ = ['download_with_tqdm',
           'tqdm_joblib']


def download_with_tqdm(url: str) -> BytesIO:
    """download url with tqdm bar
    Use in large file downloading,

    :param url: download URL
    :return: BytesIO
    """
    resp = requests.get(url, stream=True)
    resp.raise_for_status()
    file_size = int(resp.headers.get('content-length', 0))
    progress_bar = tqdm(total=file_size, unit='iB', unit_scale=True, desc='downloading from url...')
    content_stream = BytesIO()

    for data in resp.iter_content(chunk_size=1024):
        content_stream.write(data)
        progress_bar.update(len(data))

    progress_bar.close()
    content_stream.seek(0)

    return content_stream


@contextlib.contextmanager
def tqdm_joblib(tqdm_obj: tqdm) -> ContextManager[tqdm]:
    """Context manager to patch joblib multiprocessing to report into tqdm progress bar given as argument

    Example of running foreach neuron shuffle:

    .. code-block:: python

        from joblib import Parallel, delayed

        func = ... # Callable[*args, Any], arg include foreach *n*

        with tqdm_joblib(tqdm(desc="lower bound", unit='neuron', ncols=80)) as _:
            Parallel(n_jobs=self.parallel_jobs, backend='multiprocessing', verbose=True)(
                delayed(func)
                (*args)
                for n in neuron_list

    :param tqdm_obj: tqdm object
    :return: tqdm context manger

    """
    import joblib

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_obj.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback

    try:
        yield tqdm_obj
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_obj.close()



import abc
from pathlib import Path
from typing import Any, Literal

import numpy as np

from neuralib.typing import PathLike

__all__ = [
    'SceneIdx',
    'DimCode',
    'AbstractScanner',
]

SceneIdx = int
"""0-base scan position/scene"""

DimCode = str
"""
Dimension string code representing the order of axes in the data array.
Common codes (subset of tifffile):
    S - scene/series
    
    T - time
    
    C - channel
    
    Z - z plane (depth)
    
    Y - image height
    
    X - image width
    
    A - samples (for RGB/A)
    
Example: 'SZCYX'
"""


class AbstractScanner(metaclass=abc.ABCMeta):
    """Abstract Base Class for confocal microscopy image data loaders.

    Provides a common interface for accessing metadata and image data
    from various confocal file formats
    """

    _filepath: Path
    _metadata: dict[str, Any]

    def __init__(self, filepath: PathLike):
        self._filepath = Path(filepath).resolve()
        self._metadata = self._load_metadata()

    def __enter__(self):
        """Enter the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context, ensuring resources are released."""
        self.close()

    def close(self):
        """
        Explicitly close any open resources associated with the scanner.
        Subclasses should override this if they open files or other resources.
        """
        pass

    @property
    def filepath(self) -> Path:
        """The absolute path to the loaded confocal file."""
        return self._filepath

    @property
    def metadata(self) -> dict[str, Any]:
        """Return the metadata dictionary parsed from the file."""
        return self._metadata

    @abc.abstractmethod
    def _load_metadata(self) -> dict[str, Any]:
        """
        Load essential metadata from the file.

        This method is called during initialization and should parse the
        minimum metadata required to fulfill the basic properties like
        n_scenes, dimensions, etc. More extensive metadata can be loaded
        lazily by ``get_full_metadata()``.

        :return: A dictionary containing essential metadata
        """
        pass

    @property
    @abc.abstractmethod
    def dimcode(self) -> DimCode:
        """Get dimension code for the image array"""
        pass

    @property
    @abc.abstractmethod
    def n_scenes(self) -> int:
        """Total number of scenes (positions/series) in the file."""
        pass

    @abc.abstractmethod
    def get_channel_names(self, scene_idx: SceneIdx) -> list[str]:
        """
        Get the names of the fluorescence channels for a specific scene.

        :param scene_idx: The 0-based index of the scene.
        :return: A list of strings representing the channel names. Returns
            generic names (e.g., ['Channel 1', 'Channel 2']) if specific
            names are not available.
        :raise IndexError: If scene_idx is out of bounds.
        """
        pass

    @abc.abstractmethod
    def view(self, **kwargs) -> np.ndarray:
        """
        view the 2D imaging array.
        :return: A NumPy image array for view. `Array[Any, [X, Y]]`
        """
        pass

    @staticmethod
    def z_projection(stacks: np.ndarray,
                     project_type: Literal['avg', 'max', 'min', 'std', 'median'] = 'max',
                     axis: int = 0) -> np.ndarray:
        """
        Computes a z-projection of a stack of images along a specified axis using
        a chosen method of projection. The function provides flexibility to select
        different projection types such as average, maximum, minimum, standard
        deviation, or median. The default projection type is `max`.

        :param stacks: A multidimensional NumPy array representing the stack of images for z-projection.
        :param project_type: The type of projection to perform. Accepted values are:
            'avg', 'max', 'min', 'std', or 'median'. The default is 'max'.
        :param axis: The axis along which the projection will be calculated.
            This is generally set according to the stack structure. Default is 0.
        :return: A NumPy array containing the result of the z-projection along the specified axis.
        :raises ValueError: If the given project_type is not one of the accepted values.
        """
        match project_type:
            case 'avg':
                return np.mean(stacks, axis=0)
            case 'max':
                return np.max(stacks, axis=0)
            case 'min':
                return np.min(stacks, axis=0)
            case 'std':
                return np.std(stacks, axis=0)
            case 'median':
                return np.median(stacks, axis=axis)
            case _:
                raise ValueError(f'unknown project type:{project_type}')



from __future__ import annotations

import pickle
from typing import TypedDict

import h5py
import numpy as np
import polars as pl
from typing_extensions import Self

from neuralib.typing import PathLike
from neuralib.util.dataframe import DataFrameWrapper
from neuralib.util.utils import uglob

__all__ = [
    'read_facemap',
    'KeyPoint',
    'FaceMapResult',
    'KeyPointDataFrame',
    'SVDVariables',
    'KeyPointsMeta',
    'PupilDict',
    'RoiDict'
]


def read_facemap(directory: PathLike) -> FaceMapResult:
    """loading facemap result

    :param directory: facemap output directory
    """
    return FaceMapResult.from_directory(directory)


KeyPoint = str
"""keypoint name"""


class FaceMapResult:
    """Facemap result container"""

    def __init__(self, svd: SVDVariables | None,
                 meta: KeyPointsMeta | None,
                 data: h5py.Group | None,
                 with_keypoints: bool):
        """
        :param svd: attr:`~neuralib.tracking.facemap.core.SVDVariables`
        :param meta: attr:`~neuralib.tracking.facemap.core.KeyPointsMeta`
        :param data: facemap data result
        :param with_keypoints: whether it has keypoint tracking
        """
        self.svd = svd
        self.meta = meta
        self.data = data

        self._with_keypoints = with_keypoints

    @classmethod
    def from_directory(cls, directory: PathLike) -> Self:
        """
        init class loading from a directory

        :param directory: Facemap output directory
        """
        # svd
        try:
            svd_path = uglob(directory, '*.npy')
        except FileNotFoundError:
            svd = None
        else:
            svd = np.load(svd_path, allow_pickle=True).item()

        # meta
        try:
            meta_path = uglob(directory, '*.pkl')
        except FileNotFoundError:
            meta = None
            data = None
            keypoints = False
        else:
            with open(meta_path, 'rb') as f:
                meta = pickle.load(f)

            data_path = uglob(directory, '*.h5')
            data = h5py.File(data_path)['Facemap']
            keypoints = True

        return cls(svd, meta, data, keypoints)

    @property
    def with_keypoint(self) -> bool:
        return self._with_keypoints

    # ============== #
    # Pupil Tracking #
    # ============== #

    def get_pupil(self) -> PupilDict:
        """pupil tracking result

        :raises RuntimeError: If no pupil data is available.
        """
        try:
            pupil: list[PupilDict] = self.svd['pupil']
        except KeyError:
            raise RuntimeError('no pupil data found')
        else:
            return pupil[0]

    def get_pupil_area(self) -> np.ndarray:
        """pupil area. `Array[float, F]`"""
        return self.get_pupil()['area_smooth']

    def get_pupil_center_of_mass(self) -> np.ndarray:
        """center of mass of pupil tracking. `Array[float, [F, 2]]`"""
        return self.get_pupil()['com_smooth']

    def get_pupil_location_movement(self) -> np.ndarray:
        """Calculate the Euclidean distance from the origin for each point in a 2D array. `Array[float, F]`"""
        com = self.get_pupil_center_of_mass()
        return np.sqrt(np.sum(com ** 2, axis=1))

    def get_blink(self) -> np.ndarray:
        """eye blinking array. `Array[float, F]`

        :raises RuntimeError: If no blink data is available.
        """
        try:
            ret = self.svd['blink']
        except KeyError:
            raise RuntimeError('no blink data found')
        else:
            return ret[0]

    # ========= #
    # Keypoints #
    # ========= #

    @property
    def keypoints(self) -> list[KeyPoint]:
        """list of all keypoint name"""
        return list(self.data.keys())

    def get(self, *keypoint) -> KeyPointDataFrame:
        """get keypoint(s) dataframe"""
        if len(keypoint) == 1:
            return self._get(keypoint[0])
        else:
            ret = [self._get(k).dataframe() for k in keypoint]
            return KeyPointDataFrame(pl.concat(ret))

    def _get(self, keypoint: KeyPoint) -> KeyPointDataFrame:
        x = np.array(self.data[keypoint]['x'])
        y = np.array(self.data[keypoint]['y'])
        llh = np.array(self.data[keypoint]['likelihood'])
        df = pl.DataFrame({'x': x, 'y': y, 'likelihood': llh}).with_columns(pl.lit(keypoint).alias('keypoint'))
        return KeyPointDataFrame(df)


class KeyPointDataFrame(DataFrameWrapper):
    """
    Dataframe with ``x``, ``y``, ``likelihood`` and ``keypoint`` columns ::

        ┌────────────┬────────────┬────────────┬───────────┐
        │ x          ┆ y          ┆ likelihood ┆ keypoint  │
        │ ---        ┆ ---        ┆ ---        ┆ ---       │
        │ f32        ┆ f32        ┆ f32        ┆ str       │
        ╞════════════╪════════════╪════════════╪═══════════╡
        │ 374.102081 ┆ 199.159668 ┆ 0.777443   ┆ eye(back) │
        │ 373.785919 ┆ 199.425873 ┆ 0.787424   ┆ eye(back) │
        │ 374.075867 ┆ 199.507111 ┆ 0.779713   ┆ eye(back) │
        │ 374.028473 ┆ 199.359955 ┆ 0.761724   ┆ eye(back) │
        │ 374.222382 ┆ 199.777466 ┆ 0.770329   ┆ eye(back) │
        │ …          ┆ …          ┆ …          ┆ …         │
        │ 317.318756 ┆ 285.396912 ┆ 0.596486   ┆ mouth     │
        │ 318.163696 ┆ 285.492676 ┆ 0.589684   ┆ mouth     │
        │ 317.758606 ┆ 285.560425 ┆ 0.603126   ┆ mouth     │
        │ 317.453491 ┆ 285.572235 ┆ 0.573179   ┆ mouth     │
        │ 317.976196 ┆ 285.477051 ┆ 0.58359    ┆ mouth     │
        └────────────┴────────────┴────────────┴───────────┘

    """

    def __init__(self, df: pl.DataFrame):
        self._df = df

    def __repr__(self):
        return repr(self.dataframe())

    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        if dataframe is None:
            return self._df
        else:
            return KeyPointDataFrame(dataframe)

    def to_zscore(self) -> Self:
        """
        xy to zscore

        :return:
        """
        return self.with_columns([
            ((pl.col('x') - pl.col('x').mean()) / pl.col('x').std()).alias('x'),
            ((pl.col('y') - pl.col('y').mean()) / pl.col('y').std()).alias('y'),
        ])

    def with_outlier_filter(self, filter_window: int = 15,
                            baseline_window: int = 50,
                            max_spike: int = 25,
                            max_diff: int = 25) -> Self:
        """
        with outlier filter

        :param filter_window:
        :param baseline_window:
        :param max_spike:
        :param max_diff:
        :return:
        """
        from .util import filter_outliers
        x, y = filter_outliers(np.array(self['x']), np.array(self['y']), filter_window, baseline_window, max_spike, max_diff)

        return self.dataframe(pl.DataFrame({
            'x': x,
            'y': y,
            'likelihood': self['likelihood'],
            'keypoint': self['keypoint']
        }))


class PupilDict(TypedDict):
    """
    Pupil data dict

    `Dimension parameters`:

        F: number pf frames
    """
    area: np.ndarray
    """`Array[float, F]`"""
    com: np.ndarray
    """center of maze in XY. `Array[float, [F, 2]]`"""
    axdir: np.ndarray
    """`Array[float, [F, 2, 2]]`"""
    axlen: np.ndarray
    """`Array[float, [F, 2]]`"""
    area_smooth: np.ndarray
    """`Array[float, F]`"""
    com_smooth: np.ndarray
    """`Array[float, [F, 2]]`"""


class RoiDict(TypedDict, total=False):
    """Roi Dict"""
    rind: int
    rtype: str
    iROI: int
    ivid: int
    color: tuple[float, float, float]
    yrange: np.ndarray
    xrange: np.ndarray
    saturation: float
    pupil_sigma: float
    ellipse: np.ndarray
    yrange_bin: np.ndarray
    xrange_bin: np.ndarray


class SVDVariables(TypedDict, total=False):
    """SVD output from facemap
    .. seealso:: `<http://facemap.readthedocs.io/en/stable/outputs.html#roi-and-svd-processing>`_"""
    filenames: list[str]
    save_path: str
    Ly: list[int]
    Lx: list[int]
    sbin: int
    fullSVD: bool
    save_mat: bool
    Lybin: np.ndarray
    Lxbin: np.ndarray
    sybin: np.ndarray
    sxbin: np.ndarray
    LYbin: int
    LXbin: int
    avgframe: list[np.ndarray]
    avgmotion: list[np.ndarray]
    avgframe_reshape: np.ndarray
    avgmotion_reshape: np.ndarray
    motion: list[np.ndarray]
    motSv: list[np.ndarray]
    movSv: list[np.ndarray]
    motMask: list[int]
    movMask: list[int]
    motMask_reshape: list[int]
    movMask_reshape: list[int]
    motSVD: list[np.ndarray]
    movSVD: list[np.ndarray]
    pupil: list[PupilDict]
    running: list[np.ndarray]
    blink: list[np.ndarray]
    rois: list[RoiDict]
    sy: np.ndarray
    sx: np.ndarray


class KeyPointsMeta(TypedDict):
    """ Keypoint meta
    .. seealso:: `<https://facemap.readthedocs.io/en/stable/outputs.html#keypoints-processing>`_"""
    batch_size: int
    image_size: tuple[list[int], ...]
    bbox: tuple[int, ...]
    total_frames: int
    bodyparts: list[str]
    inference_speed: float



from typing import Literal

__all__ = ['Area', 'Source', 'Channel', 'TreeLevel',
           'HEMISPHERE_TYPE', 'PLANE_TYPE']

Area = str
"""Brain area"""

Source = str
"""Name of the injection alias"""

Channel = str
"""Fluorescence channel"""

TreeLevel = int
"""Hierarchical tree merge level"""

HEMISPHERE_TYPE = Literal['ipsi', 'contra', 'both']
"""brain hemisphere"""

PLANE_TYPE = Literal['coronal', 'sagittal', 'transverse']
"""brain section (plane) type"""
